"""Collection of fixtures for facilitation test implementations
"""
import getpass
import logging
import os
from pathlib import Path
import subprocess
import pytest
from tempfile import NamedTemporaryFile
from time import sleep
from urllib.request import urlopen

from datalad_next.datasets import Dataset
from datalad_next.runners import (
    call_git_lines,
    call_git_success,
)
from datalad_next.utils import patched_env
from .utils import (
    HTTPPath,
    WebDAVPath,
    assert_ssh_access,
    external_versions,
    get_git_config_global_fpath,
    md5sum,
    rmtree,
)

lgr = logging.getLogger('datalad.next.tests.fixtures')


@pytest.fixture(autouse=True, scope="session")
def reduce_logging():
    """Reduce the logging output during test runs

    DataLad emits a large amount of repetitive INFO log messages that only
    clutter the test output, and hardly ever help to identify an issue.
    This fixture modifies the standard logger to throw away all INFO level
    log messages.

    With this approach, such messages are still fed to and processes by the
    logger (in contrast to an apriori level setting).
    """
    dllgr = logging.getLogger('datalad')
    # leave a trace that this is happening
    dllgr.info("Test fixture starts suppressing INFO level messages")

    class NoInfo(logging.Filter):
        def filter(self, record):
            # it seems unnecessary to special case progress logs, moreover
            # not filtering them out will make clone/fetch/push very visible
            # in the logs with trivial messages
            #if hasattr(record, 'dlm_progress'):
            #    # this is a progress log message that may trigger something
            #    # a test is looking for
            #    return True
            if record.levelno == 20:
                # this is a plain INFO message, ignore
                return False
            else:
                return True

    noinfo = NoInfo()
    # we need to attach the filter to any handler to make it effective.
    # adding to the logger only will not effect any log messages produced
    # via descendant loggers
    for hdlr in dllgr.handlers:
        hdlr.addFilter(noinfo)


@pytest.fixture(autouse=False, scope="function")
def no_result_rendering(monkeypatch):
    """Disable datalad command result rendering for all command calls

    This is achieved by forcefully supplying `result_renderer='disabled'`
    to any command call via a patch to internal argument normalizer
    ``get_allargs_as_kwargs()``.
    """
    # we need to patch our patch function, because datalad-core's is no
    # longer used
    import datalad_next.patches.interface_utils as dnpiu

    old_get_allargs_as_kwargs = dnpiu.get_allargs_as_kwargs

    def no_render_get_allargs_as_kwargs(call, args, kwargs):
        kwargs, one, two = old_get_allargs_as_kwargs(call, args, kwargs)
        kwargs['result_renderer'] = 'disabled'
        return kwargs, one, two

    with monkeypatch.context() as m:
        m.setattr(dnpiu,
                  'get_allargs_as_kwargs',
                  no_render_get_allargs_as_kwargs)
        yield


@pytest.fixture(autouse=False, scope="function")
def tmp_keyring():
    """Patch plaintext keyring to temporarily use a different storage

    No credential read or write actions will impact any existing credential
    store of any configured backend.

    The patched backend is yielded by the fixture.
    """
    import keyring

    # the testsetup assumes this to be a plaintext backend.
    # this backend is based on a filename and maintains no state.
    # each operation opens, reads/writes, and then closes the file.
    # hence we can simply point to a different file
    backend = keyring.get_keyring()
    prev_fpath = backend.file_path

    # no tmp keyring yet, make one
    with NamedTemporaryFile(
            'r',
            prefix='datalad_tmp_keyring_',
            delete=True) as tf:
        # we must close, because windows does not like the file being open
        # already when ConfigManager would open it for reading
        tf.close()
        backend.file_path = tf.name
        with patched_env(DATALAD_TESTS_TMP_KEYRING_PATH=tf.name):
            yield backend

    backend.file_path = prev_fpath


# the following is taken from datalad/conftest.py
# sadly, this is defined inline and cannot be reused directly
standard_gitconfig = """\
[user]
    name = DataLad Tester
    email = test@example.com
[core]
    askPass =
[datalad "log"]
    exc = 1
[datalad "extensions"]
    # load the next extension to be able to test patches of annex remotes
    # that run in subprocesses
    load = next
[annex "security"]
    # from annex 6.20180626 file:/// and http://localhost access isn't
    # allowed by default
    allowed-url-schemes = http https file
    allowed-http-addresses = all
[protocol "file"]
    # since git 2.38.1 cannot by default use local clones for submodules
    # https://github.blog/2022-10-18-git-security-vulnerabilities-announced/#cve-2022-39253
    allow = always
""" + os.environ.get('DATALAD_TESTS_GITCONFIG', '').replace('\\n', os.linesep)


@pytest.fixture(autouse=False, scope="function")
def datalad_cfg():
    """Temporarily alter configuration to use a plain "global" configuration

    The global configuration manager at `datalad.cfg` is reloaded after
    adjusting `GIT_CONFIG_GLOBAL` to point to a new temporary `.gitconfig`
    file.

    After test execution the file is removed, and the global `ConfigManager`
    is reloaded once more.

    Any test using this fixture will be skipped for Git versions earlier
    than 2.32, because the `GIT_CONFIG_GLOBAL` environment variable used
    here was only introduced with that version.
    """
    if external_versions['cmd:git'] < "2.32":
        pytest.skip(
            "Git configuration redirect via GIT_CONFIG_GLOBAL "
            "only supported since Git v2.32"
        )
    from datalad import cfg
    with NamedTemporaryFile(
            'w',
            prefix='datalad_gitcfg_global_',
            delete=False) as tf:
        tf.write(standard_gitconfig)
        # we must close, because windows does not like the file being open
        # already when ConfigManager would open it for reading
        tf.close()
        with patched_env(GIT_CONFIG_GLOBAL=tf.name):
            cfg.reload(force=True)
            yield cfg
    # reload to put the previous config in effect again
    cfg.reload(force=True)


@pytest.fixture(autouse=True, scope="function")
def check_gitconfig_global():
    """No test must modify a user's global Git config.

    If such modifications are needed, a custom configuration setup
    limited to the scope of the test requiring it must be arranged.
    """
    globalcfg_fname = get_git_config_global_fpath()
    if not globalcfg_fname.exists():
        lgr.warning(
            'No global/user Git config file exists. This is an unexpected '
            'test environment, no config modifications checks can be '
            'performed. Proceeding nevertheless.')
        # let the test run
        yield
        # and exit quietly
        return

    # we have a config file. hash it pre and post test. Fail is changed.
    pre = md5sum(globalcfg_fname)
    yield
    post = md5sum(globalcfg_fname)
    assert pre == post, \
        "Global Git config modification detected. Test must be modified to use " \
        "a temporary configuration target. Hint: use the `datalad_cfg` fixture."


@pytest.fixture(autouse=True, scope="function")
def check_plaintext_keyring():
    """No test must modify a user's keyring.

    If such modifications are needed, a custom keyring setup
    limited to the scope of the test requiring it must be arranged.
    The ``tmp_keyring`` fixture can be employed in such cases.
    """
    # datalad-core configures keyring to use a plaintext backend
    # we will look for the underlying file and verify that it is either
    # no there, or remains unmodified
    import keyring
    kr = keyring.get_keyring()
    if not hasattr(kr, 'file_path'):
        # this is not the plain text keyring, nothing we can do here
        # run as-is, but leave a message
        lgr.warning('Running without the expected plain-text keyring')
        yield
        return

    kr_fpath = Path(kr.file_path)
    pre = md5sum(kr_fpath) if kr_fpath.exists() else ''
    yield
    post = md5sum(kr_fpath) if kr_fpath.exists() else ''
    assert pre == post, \
        "Keyring modification detected. Test must be modified to use " \
        "a temporary keyring. Hint: use the `tmp_keyring` fixture."


@pytest.fixture(autouse=False, scope="function")
def credman(datalad_cfg, tmp_keyring):
    """Provides a temporary credential manager

    It comes with a temporary global datalad config and a temporary
    keyring as well.

    This manager can be used to deploy or manipulate credentials within the
    scope of a single test.
    """
    from datalad import cfg
    from datalad_next.credman import CredentialManager
    cm = CredentialManager(cfg)
    yield cm


@pytest.fixture(autouse=False, scope="function")
def dataset(datalad_cfg, tmp_path_factory):
    """Provides a ``Dataset`` instance for a not-yet-existing repository

    The instance points to an existing temporary path, but ``create()``
    has not been called on it yet.
    """
    # must use the factory to get a unique path even when a concrete
    # test also uses `tmp_path`
    ds = Dataset(tmp_path_factory.mktemp("dataset"))
    yield ds


@pytest.fixture(autouse=False, scope="function")
def existing_dataset(dataset):
    """Provides a ``Dataset`` instance pointing to an existing dataset/repo

    This fixture uses an instance provided by the ``dataset`` fixture and
    calls ``create()`` on it, before it yields the ``Dataset`` instance.
    """
    dataset.create(result_renderer='disabled')
    yield dataset


@pytest.fixture(autouse=False, scope="function")
def existing_noannex_dataset(dataset):
    """just like ``existing_dataset``, but created with ``annex=False``
    """
    dataset.create(annex=False, result_renderer='disabled')
    yield dataset


@pytest.fixture(scope="session")
def modified_dataset(tmp_path_factory):
    """Produces a dataset with various modifications

    The fixture is module-scope, aiming to be reused by many tests focused
    on reporting. It does not support any further modification. The fixture
    will fail, if any such modification is detected.

    ``git status`` will report::

        ❯ git status -uall
        On branch dl-test-branch
        Changes to be committed:
          (use "git restore --staged <file>..." to unstage)
                new file:   dir_m/file_a
                new file:   file_a
                new file:   file_am

        Changes not staged for commit:
          (use "git add/rm <file>..." to update what will be committed)
          (use "git restore <file>..." to discard changes in working directory)
          (commit or discard the untracked or modified content in submodules)
                deleted:    dir_d/file_d
                deleted:    dir_m/file_d
                modified:   dir_m/file_m
                deleted:    dir_sm/sm_d
                modified:   dir_sm/sm_m (modified content)
                modified:   dir_sm/sm_mu (modified content, untracked content)
                modified:   dir_sm/sm_n (new commits)
                modified:   dir_sm/sm_nm (new commits, modified content)
                modified:   dir_sm/sm_nmu (new commits, modified content, untracked content)
                modified:   dir_sm/sm_u (untracked content)
                modified:   file_am
                deleted:    file_d
                modified:   file_m

        Untracked files:
          (use "git add <file>..." to include in what will be committed)
                dir_m/dir_u/file_u
                dir_m/file_u
                dir_u/file_u
                file_u


    Suffix indicates the ought-to state (multiple possible):

    a - added
    c - clean
    d - deleted
    n - new commits
    m - modified
    u - untracked content

    Prefix indicated the item type:

    file - file
    sm - submodule
    dir - directory
    """
    ds = Dataset(tmp_path_factory.mktemp("modified_dataset"))
    ds.create(result_renderer='disabled')
    ds_dir = ds.pathobj / 'dir_m'
    ds_dir.mkdir()
    ds_dir_d = ds.pathobj / 'dir_d'
    ds_dir_d.mkdir()
    (ds_dir / 'file_m').touch()
    (ds.pathobj / 'file_m').touch()
    dirsm = ds.pathobj / 'dir_sm'
    dss = {}
    for smname in (
        'sm_d', 'sm_c', 'sm_n', 'sm_m', 'sm_nm', 'sm_u', 'sm_mu', 'sm_nmu',
        'droppedsm_c',
    ):
        sds = Dataset(dirsm / smname).create(result_renderer='disabled')
        # for the plain modification, commit the reference right here
        if smname in ('sm_m', 'sm_nm', 'sm_mu', 'sm_nmu'):
            (sds.pathobj / 'file_m').touch()
        sds.save(to_git=True, result_renderer='disabled')
        dss[smname] = sds
    # files in superdataset to be deleted
    for d in (ds_dir_d, ds_dir, ds.pathobj):
        (d / 'file_d').touch()
    dss['.'] = ds
    dss['dir'] = ds_dir
    ds.save(to_git=True, result_renderer='disabled')
    ds.drop(dirsm / 'droppedsm_c', what='datasets', reckless='availability',
            result_renderer='disabled')
    # a new commit
    for smname in ('.', 'sm_n', 'sm_nm', 'sm_nmu'):
        sds = dss[smname]
        (sds.pathobj / 'file_c').touch()
        sds.save(to_git=True, result_renderer='disabled')
    # modified file
    for smname in ('.', 'dir', 'sm_m', 'sm_nm', 'sm_mu', 'sm_nmu'):
        obj = dss[smname]
        pobj = obj.pathobj if isinstance(obj, Dataset) else obj
        (pobj / 'file_m').write_text('modify!')
    # untracked
    for smname in ('.', 'dir', 'sm_u', 'sm_mu', 'sm_nmu'):
        obj = dss[smname]
        pobj = obj.pathobj if isinstance(obj, Dataset) else obj
        (pobj / 'file_u').touch()
        (pobj / 'dirempty_u').mkdir()
        (pobj / 'dir_u').mkdir()
        (pobj / 'dir_u' / 'file_u').touch()
    # delete items
    rmtree(dss['sm_d'].pathobj)
    rmtree(ds_dir_d)
    (ds_dir / 'file_d').unlink()
    (ds.pathobj / 'file_d').unlink()
    # added items
    for smname in ('.', 'dir', 'sm_m', 'sm_nm', 'sm_mu', 'sm_nmu'):
        obj = dss[smname]
        pobj = obj.pathobj if isinstance(obj, Dataset) else obj
        (pobj / 'file_a').write_text('added')
        assert call_git_success(['add', 'file_a'], cwd=pobj)
    # added and then modified file
    file_am_obj = ds.pathobj / 'file_am'
    file_am_obj.write_text('added')
    assert call_git_success(['add', 'file_am'], cwd=ds.pathobj)
    file_am_obj.write_text('modified')

    # record git-status output as a reference
    status_start = call_git_lines(['status'], cwd=ds.pathobj)
    yield ds
    # compare with initial git-status output, if there are any
    # differences the assumptions of any consuming test could be
    # invalidated. The modifying code must be found and fixed
    assert status_start == call_git_lines(['status'], cwd=ds.pathobj), \
        "Unexpected modification of the testbed"


@pytest.fixture(autouse=False, scope="session")
def webdav_credential():
    """Provides HTTP Basic authentication credential necessary to access the
    server provided by the ``webdav_server`` fixture."""
    yield dict(
        name='dltest-my&=webdav',
        user='datalad',
        secret='secure',
        type='user_password',
    )


@pytest.fixture(autouse=False, scope="function")
def webdav_server(tmp_path_factory, webdav_credential):
    """Provides a WebDAV server, serving a temporary directory

    The fixtures yields an instance of ``WebDAVPath``, providing the
    following essential attributes:

    - ``path``: ``Path`` instance of the served temporary directory
    - ``url``: HTTP URL to access the WebDAV server

    Server access requires HTTP Basic authentication with the credential
    provided by the ``webdav_credential`` fixture.
    """
    auth = (webdav_credential['user'], webdav_credential['secret'])
    # must use the factory to get a unique path even when a concrete
    # test also uses `tmp_path`
    path = tmp_path_factory.mktemp("webdav")
    # this looks a little awkward, but is done to avoid a change in
    # WebDAVPath.
    server = WebDAVPath(path, auth=auth)
    with server as server_url:
        server.url = server_url
        yield server


@pytest.fixture(autouse=False, scope="session")
def http_credential():
    """Provides the HTTP Basic authentication credential necessary to access the
    HTTP server provided by the ``http_server_with_basicauth`` fixture."""
    yield dict(
        name='dltest-my&=http',
        user='datalad',
        secret='secure',
        type='user_password',
    )


@pytest.fixture(autouse=False, scope="function")
def http_server(tmp_path_factory):
    """Provides an HTTP server, serving a temporary directory

    The fixtures yields an instance of ``HTTPPath``, providing the
    following essential attributes:

    - ``path``: ``Path`` instance of the served temporary directory
    - ``url``: HTTP URL to access the HTTP server
    """
    # must use the factory to get a unique path even when a concrete
    # test also uses `tmp_path`
    path = tmp_path_factory.mktemp("webdav")
    server = HTTPPath(path, use_ssl=False, auth=None)
    with server:
        # overwrite path with Path object for convenience
        server.path = path
        yield server


@pytest.fixture(autouse=False, scope="function")
def http_server_with_basicauth(tmp_path_factory, http_credential):
    """Like ``http_server`` but requiring authentication via ``http_credential``
    """
    path = tmp_path_factory.mktemp("webdav")
    server = HTTPPath(
        path, use_ssl=False,
        auth=(http_credential['user'], http_credential['secret']),
    )
    with server:
        # overwrite path with Path object for convenience
        server.path = path
        yield server


@pytest.fixture(scope="session")
def httpbin_service():
    """Return canonical access URLs for the HTTPBIN service

    This fixture tries to spin up a httpbin Docker container at localhost:8765;
    if successful, it returns this URL as the 'standard' URL.  If the attempt
    fails, a URL pointing to the canonical instance is returned.

    For tests that need to have the service served via a specific
    protocol (https vs http), the corresponding URLs are returned
    too. They always point to the canonical deployment, as some
    tests require both protocols simultaneously and a local deployment
    generally won't have https.
    """
    hburl = 'http://httpbin.org'
    hbsurl = 'https://httpbin.org'
    ciurl = 'http://localhost:8765'
    if os.name == "posix":
        try:
            r = subprocess.run(
                ["docker", "run", "-d", "-p", "127.0.0.1:8765:80", "kennethreitz/httpbin"],
                check=True,
                stdout=subprocess.PIPE,
                text=True,
            )
        except (OSError, subprocess.CalledProcessError):
            lgr.warning("Failed to spin up httpbin Docker container:", exc_info=True)
            container_id = None
        else:
            container_id = r.stdout.strip()
    else:
        container_id = None
    try:
        if container_id is not None:
            # Wait for container to fully start:
            for _ in range(25):
                try:
                    urlopen(ciurl)
                except Exception:
                    sleep(1)
                else:
                    break
            else:
                raise RuntimeError("httpbin container did not start up in time")
        yield {
            "standard": ciurl if container_id is not None else hbsurl,
            "http": hburl,
            "https": hbsurl,
        }
    finally:
        if container_id is not None:
            subprocess.run(["docker", "rm", "-f", container_id], check=True)


@pytest.fixture(scope="function")
def httpbin(httpbin_service):
    """Does the same thing as ``httpbin_service``, but skips on function-scope

    ``httpbin_service`` always returns access URLs for HTTPBIN. However,
    in some cases it is simply not desirable to run a test. For example,
    the appveyor workers are more or less constantly unable to access the
    public service. This fixture is evaluated at function-scope and
    skips the test whenever any of these undesired conditions is
    detected. Otherwise it just relays ``httpbin_service``.
    """
    if os.environ.get('DATALAD_TESTS_NONETWORK'):
        pytest.skip(
            'Not running httpbin-based test: NONETWORK flag set'
        )
    if 'APPVEYOR' in os.environ and 'DEPLOY_HTTPBIN_IMAGE' not in os.environ:
        pytest.skip(
            "Not running httpbin-based test on appveyor without "
            "docker-deployed instance -- too unreliable"
        )
    yield httpbin_service


@pytest.fixture(autouse=False, scope="function")
def datalad_interactive_ui(monkeypatch):
    """Yields a UI replacement to query for operations and stage responses

    No output will be written to STDOUT/ERR by this UI.

    A standard usage pattern is to stage one or more responses, run the
    to-be-tested code, and verify that the desired user interaction
    took place::

       > datalad_interactive_ui.staged_responses.append('skip')
       > ...
       > assert ... datalad_interactive_ui.log
    """
    from datalad_next.uis import ui_switcher
    from .utils import InteractiveTestUI

    with monkeypatch.context() as m:
        m.setattr(ui_switcher, '_ui', InteractiveTestUI())
        yield ui_switcher.ui


@pytest.fixture(autouse=False, scope="function")
def datalad_noninteractive_ui(monkeypatch):
    """Yields a UI replacement to query for operations

    No output will be written to STDOUT/ERR by this UI.

    A standard usage pattern is to run the to-be-tested code, and verify that
    the desired user messaging took place::

       > ...
       > assert ... datalad_interactive_ui.log
    """
    from datalad_next.uis import ui_switcher
    from .utils import TestUI

    with monkeypatch.context() as m:
        m.setattr(ui_switcher, '_ui', TestUI())
        yield ui_switcher.ui


@pytest.fixture(autouse=False, scope="session")
def sshserver_setup(tmp_path_factory):
    if not os.environ.get('DATALAD_TESTS_SSH'):
        pytest.skip(
            "set DATALAD_TESTS_SSH=1 to enable")

    # query a bunch of recognized configuration environment variables,
    # fill in the blanks, then check if the given configuration is working,
    # and post the full configuration again as ENV vars, to be picked up by
    # the function-scope `datalad_cfg`
    tmp_root = str(tmp_path_factory.mktemp("sshroot"))
    host = os.environ.get('DATALAD_TESTS_SERVER_SSH_HOST', 'localhost')
    port = os.environ.get('DATALAD_TESTS_SERVER_SSH_PORT', '22')
    login = os.environ.get(
        'DATALAD_TESTS_SERVER_SSH_LOGIN',
        getpass.getuser())
    seckey = os.environ.get(
        'DATALAD_TESTS_SERVER_SSH_SECKEY',
        str(Path.home() / '.ssh' / 'id_rsa'))
    path = os.environ.get('DATALAD_TESTS_SERVER_SSH_PATH', tmp_root)
    # TODO this should not use `tmp_root` unconditionally, but only if
    # the SSH_PATH is known to be the same. This might not be if SSH_PATH
    # is explicitly configured and LOCALPATH is not -- which could be
    # an indication that there is none
    localpath = os.environ.get('DATALAD_TESTS_SERVER_LOCALPATH', tmp_root)

    assert_ssh_access(host, port, login, seckey, path, localpath)

    info = {}
    # as far as we can tell, this is good, post effective config in ENV
    for v, e in (
            (host, 'HOST'),
            # this is SSH_*, because elsewhere we also have other properties
            # for other services
            (port, 'SSH_PORT'),
            (login, 'SSH_LOGIN'),
            (seckey, 'SSH_SECKEY'),
            (path, 'SSH_PATH'),
            (localpath, 'LOCALPATH'),
    ):
        os.environ[f"DATALAD_TESTS_SERVER_{e}"] = v
        info[e] = v

    yield info


@pytest.fixture(autouse=False, scope="function")
def sshserver(sshserver_setup, datalad_cfg, monkeypatch):
    # strip any leading / from the path, we add one, and
    # only one below
    sshserver_path = sshserver_setup['SSH_PATH'].lstrip('/')
    baseurl = f"ssh://{sshserver_setup['SSH_LOGIN']}" \
        f"@{sshserver_setup['HOST']}" \
        f":{sshserver_setup['SSH_PORT']}" \
        f"/{sshserver_path}"
    with monkeypatch.context() as m:
        m.setenv("DATALAD_SSH_IDENTITYFILE", sshserver_setup['SSH_SECKEY'])
        # force reload the config manager, to ensure the private key setting
        # makes it into the active config
        datalad_cfg.reload(force=True)
        yield baseurl, Path(sshserver_setup['LOCALPATH'])



import io
import stat

import pytest

from datalad_next.tests import (
    skip_if_on_windows,
)
from ..ssh import (
    SshUrlOperations,
    UrlOperationsRemoteError,
    UrlOperationsResourceUnknown,
)


# path magic inside the test is posix only
@skip_if_on_windows
def test_ssh_url_download(tmp_path, monkeypatch, sshserver):
    ssh_url, ssh_localpath = sshserver
    test_path = ssh_localpath / 'myfile'
    test_url = f'{ssh_url}/myfile'
    ops = SshUrlOperations()
    # no target file (yet), precise exception
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.stat(test_url)
    # this is different for a general connection error
    with pytest.raises(UrlOperationsRemoteError):
        ops.stat(f'ssh://localhostnotaround{test_path}')
    # now put something at the target location
    test_path.write_text('surprise!')
    # and now it works
    props = ops.stat(test_url)
    # we get the correct file size reported
    assert props['content-length'] == test_path.stat().st_size

    # and download
    download_path = tmp_path / 'download'
    props = ops.download(test_url, download_path, hash=['sha256'])
    assert props['sha256'] == '71de4622cf536ed4aa9b65fc3701f4fc5a198ace2fa0bda234fd71924267f696'
    assert props['content-length'] == 9 == test_path.stat().st_size

    # remove source and try again
    test_path.unlink()
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.download(test_url, download_path)
    # this is different for a general connection error
    with pytest.raises(UrlOperationsRemoteError):
        ops.download(f'ssh://localhostnotaround{test_path}', download_path)


# path magic inside the test is posix only
@skip_if_on_windows
def test_ssh_url_upload(tmp_path, monkeypatch, sshserver):
    ssh_url, ssh_localpath = sshserver
    payload = 'surprise!'
    payload_path = tmp_path / 'payload'
    upload_path = ssh_localpath / 'subdir' / 'myfile'
    upload_url = f'{ssh_url}/subdir/myfile'
    ops = SshUrlOperations()

    # standard error if local source is not around
    with pytest.raises(FileNotFoundError):
        ops.upload(payload_path, upload_url)

    payload_path.write_text(payload)
    # upload creates parent dirs, so the next just works.
    # this may seem strange for SSH, but FILE does it too.
    # likewise an HTTP upload is also not required to establish
    # server-side preconditions first.
    # this functionality is not about exposing a full
    # remote FS abstraction -- just upload
    ops.upload(payload_path, upload_url)
    assert upload_path.read_text() == payload


def test_ssh_url_upload_from_stdin(tmp_path, monkeypatch, sshserver):
    ssh_url, ssh_localpath = sshserver
    payload = 'surprise!'
    upload_path = ssh_localpath / 'uploaded.txt'
    upload_url = f'{ssh_url}/uploaded.txt'
    ops = SshUrlOperations()

    class StdinBufferMock:
        def __init__(self, byte_stream: bytes):
            self.buffer = io.BytesIO(byte_stream)

    with monkeypatch.context() as mp_ctx:
        mp_ctx.setattr('sys.stdin', StdinBufferMock(payload.encode()))
        ops.upload(None, upload_url)

    assert upload_path.exists()
    assert upload_path.read_text() == payload


def test_ssh_url_upload_timeout(tmp_path, monkeypatch):
    upload_url = f'ssh://localhost/not_used.txt'
    ssh_url_ops = SshUrlOperations()

    payload = 'surprise!'
    payload_path = tmp_path / 'payload'
    payload_path.write_text(payload)

    def mocked_ssh_shell_for(*args, **kwargs):

        class XShell:
            def __init__(self, *args, **kwargs):
                self.returncode = 0

            def start(self, *args, **kwargs):
                pass

        return XShell()

    with monkeypatch.context() as mp_ctx:
        import datalad_next.url_operations.ssh
        mp_ctx.setattr(
            datalad_next.url_operations.ssh.SshUrlOperations,
            'ssh_shell_for',
            mocked_ssh_shell_for
        )
        mp_ctx.setattr(datalad_next.url_operations.ssh, 'COPY_BUFSIZE', 1)
        with pytest.raises(TimeoutError):
            ssh_url_ops.upload(payload_path, upload_url, timeout=1)


def test_check_return_code():
    SshUrlOperations._check_return_code(0, 'test-0')
    with pytest.raises(UrlOperationsResourceUnknown):
        SshUrlOperations._check_return_code(244, 'test-244')
    with pytest.raises(UrlOperationsRemoteError):
        SshUrlOperations._check_return_code(None, 'test-None')
    with pytest.raises(UrlOperationsRemoteError):
        SshUrlOperations._check_return_code(1, 'test-1')


def test_ssh_stat(sshserver):
    ssh_url, ssh_local_path = sshserver

    test_path = ssh_local_path / 'file1.txt'
    test_path.write_text('content')
    test_url = f'{ssh_url}/file1.txt'

    ops = SshUrlOperations()
    props = ops.stat(test_url)
    assert props['content-length'] == test_path.stat().st_size

    test_path.unlink()
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.stat(test_url)


@skip_if_on_windows
def test_ssh_delete(sshserver):
    ssh_url, ssh_local_path = sshserver

    test_path = ssh_local_path / 'file1.txt'
    test_path.write_text('content')
    test_path.chmod(stat.S_IRUSR)

    test_dir_path = ssh_local_path / 'dir1'
    test_dir_path.mkdir()
    test_in_dir_file = test_dir_path / 'file2.txt'
    test_in_dir_file.write_text('content 2')
    test_in_dir_file.chmod(stat.S_IRUSR)
    test_dir_path.chmod(stat.S_IRUSR | stat.S_IXUSR)

    test_url = f'{ssh_url}/file1.txt'
    test_dir_url = f'{ssh_url}/dir1'

    ops = SshUrlOperations()
    ops.delete(test_url)
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.delete(test_url)

    ops.delete(test_dir_url)
    with pytest.raises(UrlOperationsResourceUnknown):
        ops.delete(test_dir_url)



"""Make push avoid refspec handling for special remote push targets

This change introduces a replacement for core's ``push.py:_push()``
with a more intelligible flow. It replaces the stalled
https://github.com/datalad/datalad/pull/6666

Importantly, it makes one behavior change, which is desirable IMHO.
Instead of rejecting to git-push any refspec for a repo with a detached
HEAD, it will attempt to push a git-annex branch for an AnnexRepo. The
respective test that ensured this behavior beyond the particular
conditions the original problem occurred in was adjusted accordingly.

All ``push`` tests from core are imported and executed to ensure
proper functioning.

Summary of the original commits patching the core implementation.

- Consolidate publication dependency handling in one place
- Consolidate tracking of git-push-dryrun exec
  Make a failed attempt discriminable from no prior attempt.
- Factor out helper to determine refspecs-to-push for a target
- Consolidate more handling of git-pushed and make conditional on an
  actual git-remote target
  This change is breaking behavior, because previously a source repository
  without an active branch would have been rejected for a push attempt.
  However, this is a bit questionable, because the git-annex branch
  might well need a push.
- Simplify push-logic: no need for a fetch, if there is no git-push
- Factor out helper to sync a remote annex-branch
- Adjust test to constrain the evaluated conditions (replacement
  tests is included here)
  As per the reasoning recorded in datalad#1811 (comment)
  the test ensuring the continue fix of datalad#1811 is actually
  verifying a situation that is not fully desirable. It prevents
  pushing of thew 'git-annex' branch whenever a repo is on a detached
  HEAD.
  This change let's the test run on a plain Git repo, where there is
  indeed nothing to push in this case.
"""

from itertools import chain
import logging
import re

import datalad.core.distributed.push as mod_push
from datalad_next.utils import (
    ensure_list,
    log_progress,
)
from datalad_next.datasets import (
    LegacyAnnexRepo as AnnexRepo,
    Dataset,
)
from . import apply_patch


lgr = logging.getLogger('datalad.core.distributed.push')


def _push(dspath, content, target, data, force, jobs, res_kwargs, pbars,
          got_path_arg=False):
    force_git_push = force in ('all', 'gitpush')

    # nothing recursive in here, we only need a repo to work with
    ds = Dataset(dspath)
    repo = ds.repo

    res_kwargs.update(type='dataset', path=dspath)

    # content will be unique for every push (even on the same dataset)
    pbar_id = 'push-{}-{}'.format(target, id(content))
    # register for final orderly take down
    pbars[pbar_id] = ds
    log_progress(
        lgr.info, pbar_id,
        'Determine push target',
        unit=' Steps',
        label='Push',
        total=4,
    )

    #
    # First we must figure out where to push to, if needed
    #

    # will contain info to determine what refspecs need to be pushed
    # and to which remote, if none is given
    wannabe_gitpush = None
    # pristine input arg
    _target = target
    # verified or auto-detected target sibling name
    target, status, message, wannabe_gitpush = _get_push_target(repo, target)
    if target is None:
        yield dict(
            res_kwargs,
            status=status,
            message=message,
        )
        return

    log_progress(
        lgr.info, pbar_id, "Push refspecs",
        label="Push to '{}'".format(target), update=1, total=4)

    # cache repo type
    is_annex_repo = isinstance(ds.repo, AnnexRepo)
    # handling pure special remotes is a lot simpler
    target_is_git_remote = repo.config.get(
        f'remote.{target}.url', None) is not None
    # TODO would is be useful to also check whether the
    # target is set 'annex-ignore' right here?

    if target_is_git_remote:
        # branch and refspec only need handling for Git remotes
        refspecs2push = _get_refspecs2push(
            repo, is_annex_repo, target, target_arg=_target,
            wannabe_gitpush=wannabe_gitpush)

        if not refspecs2push:
            # nothing was set up for push, and we have no active branch
            # this is a weird one, let's confess and stop here
            # I don't think we need to support such a scenario
            yield dict(
                res_kwargs,
                status='impossible',
                message=
                'There is no active branch, cannot determine remote '
                'branch'
            )
            return

    #
    # We know where to push to, honor dependencies
    # XXX we could do this right after we know the value of `target`,
    # but this would mean we would also push to dependencies
    # even when no actual push to the primary target is needed
    #

    # list of remotes that are publication dependencies for the
    # target remote
    # multiple dependencies could come from multiple declarations
    # of such a config items, but each declaration could also
    # contain a white-space separated list of remote names
    # see https://github.com/datalad/datalad/issues/6867
    publish_depends = list(chain.from_iterable(
        d.split() for d in ensure_list(
            ds.config.get(
                f'remote.{target}.datalad-publish-depends',
                [],
                get_all=True))))
    if publish_depends:
        lgr.debug("Discovered publication dependencies for '%s': %s'",
                  target, publish_depends)

    # we know what to push and where, now dependency processing first
    for r in publish_depends:
        # simply make a call to this function again, all the same, but
        # target is different
        # TODO: what if a publication dependency doesn't have any of the
        # determined refspecs2push, yet. Should we not attempt to push them,
        # because the main target has it?
        yield from _push(
            dspath,
            content,
            # to this particular dependency
            r,
            data,
            force,
            jobs,
            res_kwargs.copy(),
            pbars,
            got_path_arg=got_path_arg,
        )

    # and lastly the primary push target

    # git-annex data copy
    #
    if is_annex_repo:
        if data != "nothing":
            log_progress(
                lgr.info, pbar_id, "Transfer data",
                label="Transfer data to '{}'".format(target), update=2, total=4)
            yield from mod_push._transfer_data(
                repo,
                ds,
                target,
                content,
                data,
                force,
                jobs,
                res_kwargs.copy(),
                got_path_arg=got_path_arg,
            )
        else:
            lgr.debug("Data transfer to '%s' disabled by argument", target)
    else:
        lgr.debug("No data transfer: %s is not a git annex repository", repo)

    if not target_is_git_remote or not refspecs2push:
        # there is nothing that we need to push or sync with on the git-side
        # of things with this remote
        lgr.debug('No git-remote or no refspecs found that need to be pushed')
        # TODO ensure progress bar is ended properly
        return

    log_progress(
        lgr.info, pbar_id, "Update availability information",
        label="Update availability for '{}'".format(target), update=3, total=4)

    # TODO fetch is only needed if anything was actually transferred. Collect this
    # info and make the following conditional on it

    # after file transfer the remote might have different commits to
    # the annex branch. They have to be merged locally, otherwise a
    # push of it further down will fail
    _sync_remote_annex_branch(repo, target, is_annex_repo)

    # and push all relevant branches, plus the git-annex branch to announce
    # local availability info too
    yield from mod_push._push_refspecs(
        repo,
        target,
        refspecs2push,
        force_git_push,
        res_kwargs.copy(),
    )


def _append_branch_to_refspec_if_needed(repo, refspecs, branch):
    # try to anticipate any flavor of an idea of a branch ending up in a
    # refspec
    looks_like_that_branch = re.compile(
        r'((^|.*:)refs/heads/|.*:|^){}$'.format(branch))
    if all(not looks_like_that_branch.match(r) for r in refspecs):
        refspecs.append(
            branch
            if repo.config.get('branch.{}.merge'.format(branch), None)
            else '{branch}:{branch}'.format(branch=branch)
        )


def _get_push_dryrun(repo, remote=None):
    """
    Returns
    -------
    list
      The result of the dry-run. Will be an empty list if the dry-run
      failed for any reason.
    """
    try:
        wannabe_gitpush = repo.push(remote=remote, git_options=['--dry-run'])
    except Exception as e:
        lgr.debug(
            'Dry-run push to %r remote failed, '
            'assume no configuration: %s',
            remote if remote else 'default',
            e)
        wannabe_gitpush = []
    return wannabe_gitpush


def _get_push_target(repo, target_arg):
    """
    Returns
    -------
    str or None, str, str or None, list or None
      Target label, if determined; status label; optional message;
      git-push-dryrun result for reuse or None, if no dry-run was
      attempted.
    """
    # verified or auto-detected
    target = None
    # for reuse
    wannabe_gitpush = None
    if not target_arg:
        # let Git figure out what needs doing
        # we will reuse the result further down again, so nothing is wasted
        wannabe_gitpush = _get_push_dryrun(repo)
        # we did not get an explicit push target, get it from Git
        target = set(p.get('remote', None) for p in wannabe_gitpush)
        # handle case where a pushinfo record did not have a 'remote'
        # property -- should not happen, but be robust
        target.discard(None)
        if not len(target):
            return (
                None,
                'impossible',
                'No push target given, and none could be '
                'auto-detected, please specify via --to',
                wannabe_gitpush,
            )
        elif len(target) > 1:
            # dunno if this can ever happen, but if it does, report
            # nicely
            return (
                None,
                'error',
                ('No push target given, '
                 'multiple candidates auto-detected: %s',
                 list(target)),
                wannabe_gitpush,
            )
        else:
            # can only be a single one at this point
            target = target.pop()

    if not target:
        if target_arg not in repo.get_remotes():
            return (
                None,
                'error',
                ("Unknown target sibling '%s'.", target_arg),
                wannabe_gitpush,
            )
        target = target_arg

    # we must have a valid target label now
    assert target

    return (target, 'ok', None, wannabe_gitpush)


def _get_refspecs2push(repo, is_annex_repo, target, target_arg=None,
                       wannabe_gitpush=None):
    """Determine which refspecs shall be pushed to target

    Parameters
    ----------
    repo: Repo
    target: str
      Pre-determined push target
    target_arg: str, optional
      Target level given to original push() call, if any.
    wannabe_gitpush: list, optional
      Any cashed git-push-dryrun results for `target`

    Returns
    -------
    list
      Refspec labels
    """
    # (possibly redo) a push attempt to figure out what needs pushing
    # do this on the main target only, and apply the result to all
    # dependencies
    if target_arg and wannabe_gitpush is None:
        # only do it when an explicit target was given, otherwise
        # we can reuse the result from the auto-probing above
        wannabe_gitpush = _get_push_dryrun(repo, remote=target)

    refspecs2push = [
        # if an upstream branch is set, go with it
        p['from_ref']
        if repo.config.get(
            # refs come in as refs/heads/<branchname>
            # need to cut the prefix
            'branch.{}.remote'.format(p['from_ref'][11:]),
            None) == target and repo.config.get(
                'branch.{}.merge'.format(p['from_ref'][11:]),
                None)
        # if not, define target refspec explicitly to avoid having to
        # set an upstream branch, which would happen implicitly from
        # a users POV, and may also be hard to decide when publication
        # dependencies are present
        else '{}:{}'.format(p['from_ref'], p['to_ref'])
        for p in wannabe_gitpush
        if 'uptodate' not in p['operations'] and (
            # cannot think of a scenario where we would want to push a
            # managed branch directly, instead of the corresponding branch
            'refs/heads/adjusted' not in p['from_ref'])
    ]

    active_branch = repo.get_active_branch()
    if active_branch and is_annex_repo:
        # we could face a managed branch, in which case we need to
        # determine the actual one and make sure it is sync'ed with the
        # managed one, and push that one instead. following methods can
        # be called unconditionally
        repo.localsync(managed_only=True)
        active_branch = repo.get_corresponding_branch(
            active_branch) or active_branch

    # make sure that we always push the active branch (the context for the
    # potential path arguments) and the annex branch -- because we claim
    # to know better than any git config
    must_have_branches = [active_branch] if active_branch else []
    if is_annex_repo:
        must_have_branches.append('git-annex')
    for branch in must_have_branches:
        # refspecs2push= (in-place modification inside)
        _append_branch_to_refspec_if_needed(repo, refspecs2push, branch)

    return refspecs2push


def _sync_remote_annex_branch(repo, target, is_annex_repo):
    """Fetch remote annex-branch and merge locally

    Useful to ensure a push to the target will not fail due to unmerged
    remote changes.

    Parameters
    ----------
    repo: Repo
    target: str
    is_annex_repo: bool
    """
    try:
        # fetch remote, let annex sync them locally, so that the push
        # later on works.
        # We have to fetch via the push url (if there is any),
        # not a pull url.
        # The latter might be dumb and without the execution of a
        # post-update hook we might not be able to retrieve the
        # server-side git-annex branch updates (and git-annex does
        # not trigger the hook on copy), but we know we have
        # full access via the push url -- we have just used it to copy.
        lgr.debug("Fetching 'git-annex' branch updates from '%s'", target)
        fetch_cmd = ['fetch', target, 'git-annex']
        pushurl = repo.config.get(
            'remote.{}.pushurl'.format(target), None)
        if pushurl:
            # for some reason overwriting remote.{target}.url
            # does not have any effect...
            fetch_cmd = [
                '-c',
                'url.{}.insteadof={}'.format(
                    pushurl,
                    repo.config.get(
                        'remote.{}.url'.format(target), None)
                )
            ] + fetch_cmd
            lgr.debug(
                "Sync local annex branch from pushurl after remote "
                'availability update.')
        # XXX when this is changed to `call_git()`,
        # make sure to `force_c_locale=True`
        repo.call_git(fetch_cmd)
        # If no CommandError was raised, it means that remote has git-annex
        # but local repo might not be an annex yet. Since there is nothing to "sync"
        # from us, we just skip localsync without mutating repo into an AnnexRepo
        if is_annex_repo:
            repo.localsync(target)
    except mod_push.CommandError as e:
        # it is OK if the remote doesn't have a git-annex branch yet
        # (e.g. fresh repo)
        # Is this even possible? we just copied? Maybe check if anything
        # was actually copied?
        # Yes, this is possible. The current implementation of the datalad-annex
        # special remote would run into this situation. It would copy annex objects
        # to a new location just fine, but until a repository deposit was made
        # (and this implementation of push only does this as a second step), it
        # could not retrieve any refs from the remote.
        # the following conditional tests for the common prefix of the respective
        # error message by Git and the Git-channeled error message from the
        # datalad-annex remote helper.
        if "fatal: couldn't find remote ref" not in e.stderr.lower():
            raise
        lgr.debug('Remote does not have a git-annex branch: %s', e)


apply_patch('datalad.core.distributed.push', None, '_push', _push)



"""Small helpers for HTTP operations
"""
# allow for |-type UnionType declarations
from __future__ import annotations

import logging
from urllib.parse import urlparse


lgr = logging.getLogger('datalad.ext.next.utils.http_helpers')


__all__ = ['get_auth_realm']


def get_auth_realm(url, auth_info, scheme=None):
    """Determine an authentication realm identifier from a HTTP response.

    Examples
    --------
    Robustly determine a realm identifier for any URL::

       > url, props = HttpUrlOperations().probe_url(
             'https://fz-juelich.sciebo.de/...')
       > get_auth_realm(url, props.get('auth'))
       'https://fz-juelich.sciebo.de/login'

    Parameters
    ----------
    url: str
      A URL as returned by `probe_url()`
    auth_info: dict
      A mapping of supported authentication schemes to the properties
      (i.e., a 'www-authenticate' response header), as returned by
      `probe_url()`'s 'auth' property.
    scheme: str, optional
      Which specific authentication to report a realm for, in case
      multiple are supported (such as 'basic', or 'token').
      If not given, the first (if any) reported authentication
      scheme is reported on.

    Returns
    -------
    str
      A server-specific realm identifier
    """
    if not auth_info:
        # no info from the server on what it needs
        # out best bet is the URL itself
        return url
    if scheme:
        auth_info = auth_info[scheme]
    else:
        scheme, auth_info = auth_info.copy().popitem()
    # take any, but be satisfied with none too
    realm = auth_info.get('realm') if auth_info else ''
    # a realm is supposed to indicate a validity scope of a credential
    # on a server. so we make sure to have the return realm identifier
    # actually indicate a server too, in order to make it suitable for
    # a global credential lookup
    if _is_valid_url(realm):
        # the realm is already a valid URL with a server specification.
        # we can simply relay it as such, following the admins' judgement
        return realm
    else:
        # the realm was just some kind of string. we prefix it with the
        # netloc of the given URL (ignoring its path) to achieve
        # the same server-specific realm semantics
        parsed = urlparse(url)
        return '{scheme}://{netloc}{slash}{realm}'.format(
            scheme=parsed.scheme,
            netloc=parsed.netloc,
            slash='' if realm.startswith('/') else'/',
            realm=realm,
        )


def _is_valid_url(url):
    try:
        parsed = urlparse(url)
        return all([parsed.scheme, parsed.netloc])
    except:
        return False



"""
"""

from __future__ import annotations

__docformat__ = 'restructuredtext'

from dataclasses import (
    asdict,
    dataclass,
)
from datetime import datetime
from humanize import (
    naturalsize,
    naturaldate,
    naturaltime,
)
from logging import getLogger
from pathlib import Path
from stat import filemode
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
)

from datalad_next.commands import (
    EnsureCommandParameterization,
    ValidatedInterface,
    Parameter,
    ParameterConstraintContext,
    build_doc,
    eval_results,
    get_status_dict,
)
from datalad_next.constraints import (
    EnsureChoice,
    EnsurePath,
    EnsureURL,
    EnsureHashAlgorithm,
    EnsureListOf,
)
from datalad_next.uis import (
    ansi_colors as ac,
    ui_switcher as ui,
)
from datalad_next.utils import ensure_list

from datalad_next.iter_collections import (
    FileSystemItemType,
    GitTreeItemType,
    GitWorktreeFileSystemItem,
    compute_multihash_from_fp,
    iter_annexworktree,
    iter_dir,
    iter_gittree,
    iter_gitworktree,
    iter_tar,
    iter_zip,
)


lgr = getLogger('datalad.local.ls_file_collection')


# hand-maintain a list of collection type names that should be
# advertised and supported. it makes little sense to auto-discover
# them, because each collection type likely needs some custom glue
# code, and some iterators may not even be about *file* collections
_supported_collection_types = (
    'directory',
    'tarfile',
    'zipfile',
    'gittree',
    'gitworktree',
    'annexworktree',
)


@dataclass  # sadly PY3.10+ only (kw_only=True)
class CollectionSpec:
    """Internal type for passing a collection specification to
    ``ls_file_collection``. it is created by the command validator
    transparently.
    """
    orig_id: Any
    iter: Iterator
    item2res: Callable


class LsFileCollectionParamValidator(EnsureCommandParameterization):
    """Parameter validator for the ``ls_file_collection`` command"""
    _collection_types = EnsureChoice(*_supported_collection_types)

    def __init__(self):
        super().__init__(
            param_constraints=dict(
                type=self._collection_types,
                collection=EnsurePath(lexists=True) | EnsureURL(),
                hash=EnsureHashAlgorithm() | EnsureListOf(EnsureHashAlgorithm()),
            ),
            joint_constraints={
                ParameterConstraintContext(('type', 'collection', 'hash'),
                                           'collection iterator'):
                self.get_collection_iter,
            },
        )

    def get_collection_iter(self, **kwargs):
        type = kwargs['type']
        collection = kwargs['collection']
        hash = kwargs['hash']
        iter_fx = None
        iter_kwargs = None
        if type in ('directory', 'tarfile', 'zipfile', 'gitworktree', 'annexworktree'):
            if not isinstance(collection, Path):
                self.raise_for(
                    kwargs,
                    "{type} collection requires a Path-type identifier",
                    type=type,
                )
            iter_kwargs = dict(
                path=collection,
                fp=hash is not None,
            )
            item2res = fsitem_to_dict
        if type == 'directory':
            iter_fx = iter_dir
            item2res = fsitem_to_dict
        elif type == 'tarfile':
            iter_fx = iter_tar
            item2res = fsitem_to_dict
        elif type == 'zipfile':
            iter_fx = iter_zip
            item2res = fsitem_to_dict
        elif type == 'gittree':
            if hash is not None:
                self.raise_for(
                    kwargs,
                    "gittree collection does not support "
                    "content hash reporting",
                )
            iter_fx = iter_gittree
            item2res = gittreeitem_to_dict
            iter_kwargs = dict(
                path=Path('.'),
                treeish=collection,
            )
        elif type == 'gitworktree':
            iter_fx = iter_gitworktree
            item2res = gitworktreeitem_to_dict
        elif type == 'annexworktree':
            iter_fx = iter_annexworktree
            item2res = annexworktreeitem_to_dict
        else:
            raise RuntimeError(
                'unhandled collection-type: this is a defect, please report.')
        assert iter_fx is not None
        return dict(
            collection=CollectionSpec(
                orig_id=collection,
                iter=iter_fx(**iter_kwargs),
                item2res=item2res),
        )


def fsitem_to_dict(item, hash) -> Dict:
    keymap = {'name': 'item'}
    # FileSystemItemType is too fine-grained to be used as result type
    # directly, map some cases!
    fsitem_type_to_res_type = {
        'specialfile': 'file',
    }

    # file-objects need special handling (cannot be pickled for asdict())
    fp = item.fp
    item.fp = None

    # TODO likely could be faster by moving the conditional out of the
    # dict-comprehension and handling them separately upfront/after
    d = {
        keymap.get(k, k):
        # explicit str value access, until we can use `StrEnum`
        v if k != 'type' else fsitem_type_to_res_type.get(v.value, v.value)
        for k, v in asdict(item).items()
        # strip pointless symlink target reports for anything but symlinks
        if item.type is FileSystemItemType.symlink or k != 'link_target'
    }
    if fp:
        for hname, hdigest in compute_multihash_from_fp(fp, hash).items():
            d[f'hash-{hname}'] = hdigest
        # we also provide the file pointer to the consumer, although
        # it may have been "exhausted" by the hashing above and would
        # need a seek(0) for any further processing.
        # however, we do not do this here, because it is generic code,
        # and we do not know whether a particular file-like even supports
        # seek() under all circumstances. we simply document the fact.
        d['fp'] = fp
    return d


def gittreeitem_to_dict(item, hash) -> Dict:
    gittreeitem_type_to_res_type = {
        # permission bits are not distinguished for types
        GitTreeItemType.executablefile: 'file',
        # 'dataset' is the commonly used label as the command API
        # level
        GitTreeItemType.submodule: 'dataset',
    }

    gittype = gittreeitem_type_to_res_type.get(
        item.gittype, item.gittype.value) if item.gittype else None

    d = dict(item=item.name)
    if gittype is not None:
        d['type'] = gittype

    if item.gitsha:
        d['gitsha'] = item.gitsha

    if gittype is not None:
        d['gittype'] = gittype
    return d


def gitworktreeitem_to_dict(item, hash) -> Dict:
    gitworktreeitem_type_to_res_type = {
        # permission bits are not distinguished for types
        GitTreeItemType.executablefile: 'file',
        # 'dataset' is the commonly used label as the command API
        # level
        GitTreeItemType.submodule: 'dataset',
    }

    gittype = gitworktreeitem_type_to_res_type.get(
        item.gittype, item.gittype.value) if item.gittype else None

    if isinstance(item, GitWorktreeFileSystemItem):
        d = fsitem_to_dict(item, hash)
    else:
        d = dict(item=item.name)
        if gittype is not None:
            d['type'] = gittype

    if item.gitsha:
        d['gitsha'] = item.gitsha

    if gittype is not None:
        d['gittype'] = gittype
    return d


def annexworktreeitem_to_dict(item, hash) -> Dict:
    d = gitworktreeitem_to_dict(item, hash)
    if item.annexkey:
        d['type'] = 'annexed file'
        d['annexkey'] = item.annexkey
        d['annexsize'] = item.annexsize
        d['annexobjpath'] = item.annexobjpath

    return d


@build_doc
class LsFileCollection(ValidatedInterface):
    """Report information on files in a collection

    This is a utility that can be used to query information on files in
    different file collections. The type of information reported varies across
    collection types. However, each result at minimum contains some kind of
    identifier for the collection ('collection' property), and an identifier
    for the respective collection item ('item' property). Each result
    also contains a ``type`` property that indicates particular type of file
    that is being reported on. In most cases this will be ``file``, but
    other categories like ``symlink`` or ``directory`` are recognized too.

    If a collection type provides file-access, this command can compute one or
    more hashes (checksums) for any file in a collection.

    Supported file collection types are:

    ``directory``
      Reports on the content of a given directory (non-recursively). The
      collection identifier is the path of the directory. Item identifiers
      are the names of items within that directory. Standard properties like
      ``size``, ``mtime``, or ``link_target`` are included in the report.
      [PY: When hashes are computed, an ``fp`` property with a file-like
      is provided. Reading file data from it requires a ``seek(0)`` in most
      cases. This file handle is only open when items are yielded directly
      by this command (``return_type='generator``) and only until the next
      result is yielded. PY]

    ``gittree``
      Reports on the content of a Git "tree-ish". The collection identifier
      is that tree-ish. The command must be executed inside a Git repository.
      If the working directory for the command is not the repository root
      (in case of a non-bare repository), the report is constrained to
      items underneath the working directory. Item identifiers
      are the relative paths of items within that working directory.
      Reported properties include ``gitsha`` and ``gittype``; note that the
      ``gitsha`` is not equivalent to a SHA1 hash of a file's content, but
      is the SHA-type blob identifier as reported and used by Git.
      Reporting of content hashes beyond the ``gitsha`` is presently not
      supported.

    ``gitworktree``
      Reports on all tracked and untracked content of a Git repository's
      work tree. The collection identifier is a path of a directory in a Git
      repository (which can, but needs not be, its root). Item identifiers
      are the relative paths of items within that directory. Reported
      properties include ``gitsha`` and ``gittype``; note that the
      ``gitsha`` is not equivalent to a SHA1 hash of a file's content, but
      is the SHA-type blob identifier as reported and used by Git.
      [PY: When hashes are computed, an ``fp`` property with a file-like is
      provided. Reading file data from it requires a ``seek(0)`` in most
      cases. This file handle is only open when items are yielded directly
      by this command (``return_type='generator``) and only until the next
      result is yielded. PY]

    ``annexworktree``
      Like ``gitworktree``, but amends the reported items with git-annex
      information, such as ``annexkey``, ``annexsize``, and ``annnexobjpath``.

    ``tarfile``
      Reports on members of a TAR archive. The collection identifier is the
      path of the TAR file. Item identifiers are the relative paths
      of archive members within the archive. Reported properties are similar
      to the ``directory`` collection type.
      [PY: When hashes are computed, an ``fp`` property with a file-like
      is provided. Reading file data from it requires a ``seek(0)`` in most
      cases. This file handle is only open when items are yielded directly
      by this command (``return_type='generator``) and only until the next
      result is yielded. PY]

    ``zipfile``
      Like ``tarfile`` for reporting on ZIP archives.
    """
    _validator_ = LsFileCollectionParamValidator()

    # this is largely here for documentation and CLI parser building
    _params_ = dict(
        type=Parameter(
            args=("type",),
            choices=_supported_collection_types,
            doc="""Name of the type of file collection to report on"""),
        collection=Parameter(
            args=('collection',),
            metavar='ID/LOCATION',
            doc="""identifier or location of the file collection to report on.
            Depending on the type of collection to process, the specific
            nature of this parameter can be different. A common identifier
            for a file collection is a path (to a directory, to an archive),
            but might also be a URL. See the documentation for details on
            supported collection types."""),
        hash=Parameter(
            args=("--hash",),
            action='append',
            metavar='ALGORITHM',
            doc="""One or more names of algorithms to be used for reporting
            file hashes. They must be supported by the Python 'hashlib' module,
            e.g. 'md5' or 'sha256'. Reporting file hashes typically
            implies retrieving/reading file content. This processing
            may also enable reporting of additional properties that
            may otherwise not be readily available.
            [CMD: This option can be given more than once CMD]
            """),
    )

    _examples_: List = [
        {'text': 'Report on the content of a directory',
         'code_cmd': 'datalad -f json ls-file-collection directory /tmp',
         'code_py': 'records = ls_file_collection("directory", "/tmp")'},
        {'text': 'Report on the content of a TAR archive with '
                 'MD5 and SHA1 file hashes',
         'code_cmd': 'datalad -f json ls-file-collection'
                     ' --hash md5 --hash sha1 tarfile myarchive.tar.gz',
         'code_py': 'records = ls_file_collection("tarfile",'
                    ' "myarchive.tar.gz", hash=["md5", "sha1"])'},
        {'text': "Register URLs for files in a directory that is"
                 " also reachable via HTTP. This uses ``ls-file-collection``"
                 " for listing files and computing MD5 hashes,"
                 " then using ``jq`` to filter and transform the output"
                 " (just file records, and in a JSON array),"
                 " and passes them to ``addurls``, which generates"
                 " annex keys/files and assigns URLs."
                 " When the command finishes, the dataset contains no"
                 " data, but can retrieve the files after confirming"
                 " their availability (i.e., via `git annex fsck`)",
         'code_cmd':
         'datalad -f json ls-file-collection directory wwwdir --hash md5 \\\n'
         ' | jq \'. | select(.type == "file")\' \\\n'
         ' | jq --slurp . \\\n'
         " | datalad addurls --key 'et:MD5-s{size}--{hash-md5}' - 'https://example.com/{item}'"},
        {'text': 'List annex keys of all files in the working tree of a dataset',
         'code_py': "[r['annexkey'] \\\n"
                    "for r in ls_file_collection('annexworktree', '.') \\\n"
                    "if 'annexkey' in r]",
         'code_cmd': "datalad -f json ls-file-collection annexworktree . \\\n"
                     "| jq '. | select(.annexkey) | .annexkey'",
        },
    ]

    @staticmethod
    @eval_results
    def __call__(
            type: str,
            collection: CollectionSpec,
            *,
            hash: str | List[str] | None = None,
    ):
        for item in collection.iter:
            res = collection.item2res(
                item,
                hash=ensure_list(hash),
            )
            res.update(get_status_dict(
                action='ls_file_collection',
                status='ok',
                collection=collection.orig_id,
            ))
            yield res

    @staticmethod
    def custom_result_renderer(res, **kwargs):
        # given the to-be-expected diversity, this renderer only
        # outputs identifiers and type info. In almost any real use case
        # either no rendering or JSON rendering will be needed

        type = res.get('type', None)

        # if there is no mode, produces '?---------'
        # .. or 0 is needed, because some iterators report an explicit
        # `None` mode
        mode = filemode(res.get('mode', 0) or 0)

        size = None
        if type in ('file', 'hardlink'):
            size = res.get('size', None)
        size = '-' if size is None else naturalsize(size, gnu=True)

        mtime = res.get('mtime', '')
        if mtime:
            dt = datetime.fromtimestamp(mtime)
            hts = naturaldate(dt)
            if hts == 'today':
                hts = naturaltime(dt)
                hts = hts.replace(
                    'minutes ago', 'min ago').replace(
                    'seconds ago', 'sec ago')

        # stick with numerical IDs (although less accessible), we cannot
        # know in general whether this particular system can map numerical
        # IDs to valid target names (think stored name in tarballs)
        owner_info = f'{res["uid"]}:{res["gid"]}' if res.get('uid') else ''

        ui.message('{mode} {size: >6} {owner: >9} {hts: >11} {item} ({type})'.format(
            mode=mode,
            size=size,
            owner=owner_info,
            hts=hts if mtime else '',
            item=ac.color_word(
                res.get('item', '<no-item-identifier>'),
                ac.BOLD),
            type=ac.color_word(
                res.get('type', '<no-type-information>'),
                ac.MAGENTA),
        ))



"""
"""
from __future__ import annotations

__docformat__ = 'restructuredtext'

from dataclasses import dataclass
from enum import Enum
from logging import getLogger
from pathlib import Path
from typing import Generator

from datalad_next.commands import (
    CommandResult,
    CommandResultStatus,
    EnsureCommandParameterization,
    ValidatedInterface,
    Parameter,
    ParameterConstraintContext,
    build_doc,
    datasetmethod,
    eval_results,
)
from datalad_next.constraints import (
    EnsureChoice,
    EnsureDataset,
    WithDescription,
)

from datalad_next.iter_collections import (
    GitDiffStatus,
    GitTreeItemType,
    GitContainerModificationType,
    iter_gitstatus,
)
from datalad_next.uis import (
    ui_switcher as ui,
    ansi_colors as ac,
)

lgr = getLogger('datalad.core.local.status')


# TODO Could be `StrEnum`, came with PY3.11
class StatusState(Enum):
    """Enumeration of possible states of a status command result

    The "state" is the condition of the dataset item being reported
    on.
    """
    clean = 'clean'
    added = 'added'
    modified = 'modified'
    deleted = 'deleted'
    untracked = 'untracked'
    unknown = 'unknown'


diffstatus2resultstate_map = {
    GitDiffStatus.addition: StatusState.added,
    GitDiffStatus.copy: StatusState.added,
    GitDiffStatus.deletion: StatusState.deleted,
    GitDiffStatus.modification: StatusState.modified,
    GitDiffStatus.rename: StatusState.added,
    GitDiffStatus.typechange: StatusState.modified,
    GitDiffStatus.unmerged: StatusState.unknown,
    GitDiffStatus.unknown: StatusState.unknown,
    GitDiffStatus.other: StatusState.untracked,
}


# see base class decorator comment for why this is commented out
#@dataclass(kw_only=True)
@dataclass
class StatusResult(CommandResult):
    # TODO any of the following property are not actually optional
    # we only have to declare them such for limitations of dataclasses
    # prior PY3.10 (see kw_only command in base class

    diff_state: GitDiffStatus | None = None
    """The ``status`` of the underlying ``GitDiffItem``. It is named
    "_state" to emphasize the conceptual similarity with the legacy
    property 'state'
    """
    gittype: GitTreeItemType | None = None
    """The ``gittype`` of the underlying ``GitDiffItem``."""
    prev_gittype: GitTreeItemType | None = None
    """The ``prev_gittype`` of the underlying ``GitDiffItem``."""
    modification_types: tuple[GitContainerModificationType] | None = None
    """Qualifiers for modification types of container-type
    items (directories, submodules)."""

    @property
    def state(self) -> StatusState:
        """A (more or less legacy) simplified representation of the subject
        state. For a more accurate classification use the ``diff_status``
        property.
        """
        return diffstatus2resultstate_map[self.diff_state]

    # the previous status-implementation did not report plain git-types
    # we establish a getter to perform this kind of inference/mangling,
    # when it is needed
    @property
    def type(self) -> str | None:
        """
        """
        # TODO this is just a placeholder
        return self.gittype.value if self.gittype else None

    # we need a setter for this `type`-override stunt
    @type.setter
    def type(self, value):
        self.gittype = value

    @property
    def prev_type(self) -> str:
        """
        """
        return self.prev_gittype.value if self.prev_gittype else None

    @property
    def type_src(self) -> str | None:
        """Backward-compatibility adaptor"""
        return self.prev_type


opt_untracked_values = ('no', 'whole-dir', 'no-empty-dir', 'normal', 'all')
opt_recursive_values = ('no', 'repository', 'datasets', 'mono')
opt_eval_subdataset_state_values = ('no', 'commit', 'full')


class StatusParamValidator(EnsureCommandParameterization):
    def __init__(self):
        super().__init__(
            param_constraints=dict(
                # if given, it must also exist
                dataset=EnsureDataset(installed=True),
                untracked=EnsureChoice(*opt_untracked_values),
                recursive=EnsureChoice(*opt_recursive_values),
                eval_subdataset_state=EnsureChoice(
                    *opt_eval_subdataset_state_values)
            ),
            validate_defaults=('dataset',),
            joint_constraints={
                ParameterConstraintContext(('untracked', 'recursive'),
                                           'option normalization'):
                self.normalize_options,
            },
        )

    def normalize_options(self, **kwargs):
        if kwargs['untracked'] == 'no':
            kwargs['untracked'] = None
        if kwargs['untracked'] == 'normal':
            kwargs['untracked'] = 'no-empty-dir'
        if kwargs['recursive'] == 'datasets':
            kwargs['recursive'] = 'submodules'
        if kwargs['recursive'] == 'mono':
            kwargs['recursive'] = 'monolithic'
        return kwargs


@build_doc
class Status(ValidatedInterface):
    """Report on the (modification) status of a dataset

    .. note::

        This is a preview of an command implementation aiming to replace
        the DataLad ``status`` command.

        For now, expect anything here to change again.

    This command provides a report that is roughly identical to that of
    ``git status``. Running with default parameters yields a report that
    should look familiar to Git and DataLad users alike, and contain
    the same information as offered by ``git status``.

    The main difference to ``git status`` are:

    - Support for recursion into submodule. ``git status`` does that too,
      but the report is limited to the global state of an entire submodule,
      whereas this command can issue detailed reports in changes inside
      a submodule (any nesting depth).

    - Support for directory-constrained reporting. Much like ``git status``
      limits its report to a single repository, this command can optionally
      limit its report to a single directory and its direct children. In this
      report subdirectories are considered containers (much like) submodules,
      and a change summary is provided for them.

    - Support for a "mono" (monolithic repository) report. Unlike a standard
      recursion into submodules, and checking each of them for changes with
      respect to the HEAD commit of the worktree, this report compares a
      submodule with respect to the state recorded in its parent repository.
      This provides an equally comprehensive status report from the point of
      view of a queried repository, but does not include a dedicated item on
      the global state of a submodule. This makes nested hierarchy of
      repositories appear like a single (mono) repository.

    - Support for "adjusted mode" git-annex repositories. These utilize a
      managed branch that is repeatedly rewritten, hence is not suitable
      for tracking within a parent repository. Instead, the underlying
      "corresponding branch" is used, which contains the equivalent content
      in an un-adjusted form, persistently. This command detects this condition
      and automatically check a repositories state against the corresponding
      branch state.

    *Presently missing/planned features*

    - There is no support for specifying paths (or pathspecs) for constraining
      the operation to specific dataset parts. This will be added in the
      future.

    - There is no reporting of git-annex properties, such as tracked file size.
      It is undetermined whether this will be added in the future. However,
      even without a dedicated switch, this command has support for
      datasets (and their submodules) in git-annex's "adjusted mode".

    *Differences to the ``status`` command implementation prior DataLad v2*

    - Like ``git status`` this implementation reports on dataset modification,
      whereas the previous ``status`` also provided a listing of unchanged
      dataset content. This is no longer done. Equivalent functionality for
      listing dataset content is provided by the ``ls_file_collection``
      command.
    - The implementation is substantially faster. Depending on the context
      the speed-up is typically somewhere between 2x and 100x.
    - The implementation does not suffer from the limitation re type change
      detection.
    - Python and CLI API of the command use uniform parameter validation.
    """
    # Interface.validate_args() will inspect this dict for the presence of a
    # validator for particular parameters
    _validator_ = StatusParamValidator()

    # this is largely here for documentation and CLI parser building
    _params_ = dict(
        dataset=Parameter(
            args=("-d", "--dataset"),
            doc="""Dataset to be used as a configuration source. Beyond
            reading configuration items, this command does not interact with
            the dataset."""),
        untracked=Parameter(
            args=('--untracked',),
            choices=opt_untracked_values,
            doc="""Determine how untracked content is considered and reported
            when comparing a revision to the state of the working tree.
            'no': no untracked content is considered as a change;
            'normal': untracked files and entire untracked directories are
            reported as such;
            'all': report individual files even in fully untracked directories.
            In addition to these git-status modes,
            'whole-dir' (like normal, but include empty directories), and
            'no-empty-dir' (alias for 'normal') are understood."""),
        recursive=Parameter(
            args=('-r', '--recursive'),
            nargs='?',
            const='datasets',
            choices=opt_recursive_values,
            doc="""Mode of recursion for status reporting.
            With 'no' the report is restricted to a single directory and
            its direct children.
            With 'repository', the report comprises all repository content
            underneath current working directory or root of a given dataset,
            but is limited to items directly contained in that repository.
            With 'datasets', the report also comprises any content in any
            subdatasets. Each subdataset is evaluated against its respective
            HEAD commit.
            With 'mono', a report similar to 'datasets' is generated, but
            any subdataset is evaluate with respect to the state recorded
            in its parent repository. In contrast to the 'datasets' mode,
            no report items on a joint submodule are generated.
            [CMD: If no particular value is given with this option the
            'datasets' mode is selected. CMD]
            """),
        eval_subdataset_state=Parameter(
            args=("-e", "--eval-subdataset-state",),
            choices=opt_eval_subdataset_state_values,
            doc="""Evaluation of subdataset state (modified or untracked
            content) can be expensive for deep dataset hierarchies
            as subdataset have to be tested recursively for
            uncommitted modifications. Setting this option to
            'no' or 'commit' can substantially boost performance
            by limiting what is being tested.
            With 'no' no state is evaluated and subdataset are not
            investigated for modifications.
            With 'commit' only a discrepancy of the HEAD commit
            gitsha of a subdataset and the gitsha recorded in the
            superdataset's record is evaluated.
            With 'full' any other modifications are considered
            too."""),
    )

    _examples_ = [
    ]

    @staticmethod
    @datasetmethod(name="next_status")
    @eval_results
    def __call__(
        # TODO later
        #path=None,
        *,
        dataset=None,
        # TODO possibly later
        #annex=None,
        untracked='normal',
        recursive='repository',
        eval_subdataset_state='full',
    ) -> Generator[StatusResult, None, None] | list[StatusResult]:
        ds = dataset.ds
        rootpath = Path.cwd() if dataset.original is None else ds.pathobj

        for item in iter_gitstatus(
            path=rootpath,
            untracked=untracked,
            recursive=recursive,
            eval_submodule_state=eval_subdataset_state,
        ):
            yield StatusResult(
                action='status',
                status=CommandResultStatus.ok,
                path=rootpath / (item.path or item.prev_path),
                gittype=item.gittype,
                prev_gittype=item.prev_gittype,
                diff_state=item.status,
                modification_types=item.modification_types,
                refds=ds,
                logger=lgr,
            )

    def custom_result_renderer(res, **kwargs):
        # we are guaranteed to have dataset-arg info through uniform
        # parameter validation
        dsarg = kwargs['dataset']
        rootpath = Path.cwd() if dsarg.original is None else dsarg.ds.pathobj
        # because we can always determine the root path of the command
        # execution environment, we can report meaningful relative paths
        # unconditionally
        path = res.path.relative_to(rootpath)
        # collapse item type information across current and previous states
        type_ = res.type or res.prev_type or ''
        max_len = len('untracked')
        state = res.state.value
        # message format is same as for previous command implementation
        ui.message(u'{fill}{state}: {path}{type_}{annot}'.format(
            fill=' ' * max(0, max_len - len(state)),
            state=ac.color_word(
                res.state.value,
                _get_result_status_render_color(res)),
            path=path,
            type_=' ({})'.format(ac.color_word(type_, ac.MAGENTA))
            if type_ else '',
            annot=f' [{", ".join(q.value for q in res.modification_types)}]'
            if res.modification_types else '',
        ))

    @staticmethod
    def custom_result_summary_renderer(results):
        # no reports, no changes
        if len(results) == 0:
            ui.message("nothing to save, working tree clean")


def _get_result_status_render_color(res):
    if res.state == StatusState.deleted:
        return ac.RED
    elif res.state == StatusState.modified:
        return ac.CYAN
    elif res.state == StatusState.added:
        return ac.GREEN
    else:
        return ac.BOLD



# allow for |-type UnionType declarations
from __future__ import annotations

from abc import (
    ABC,
    abstractmethod,
)
from contextlib import contextmanager
from typing import (
    Any,
    Generator,
    IO,
)

import datalad

from datalad_next.config import ConfigManager
from datalad_next.iter_collections import FileSystemItem


class ArchiveOperations(ABC):
    """Base class of all archives handlers

    Any handler can be used as a context manager to adequately acquire and
    release any resources necessary to access an archive. Alternatively,
    the :func:`~ArchiveOperations.close()` method can be called, when archive
    access is no longer needed.

    In addition to the :func:`~ArchiveOperations.open()` method for accessing
    archive item content, each handler implements the standard
    ``__contains__()``, and ``__iter__()`` methods.

    ``__contains__() -> bool`` reports whether the archive contains an items of
    a given identifier.

    ``__iter__()`` provides an iterator that yields
    :class:`~datalad_next.iter_collections.FileSystemItem` instances with
    information on each archive item.
    """
    def __init__(self, location: Any, *, cfg: ConfigManager | None = None):
        """
        Parameters
        ----------
        location:
          Archive location identifier (path, URL, etc.) understood by a
          particular archive handler.
        cfg: ConfigManager, optional
          A config manager instance that implementations will consult for
          any configuration items they may support.
        """
        self._cfg = cfg
        self._location = location

    def __str__(self) -> str:
        return f'{self.__class__.__name__}({self._location})'

    def __repr__(self) -> str:
        return \
            f'{self.__class__.__name__}({self._location}, cfg={self._cfg!r})'

    @property
    def cfg(self) -> ConfigManager:
        """ConfigManager given to the constructor, or the session default"""
        if self._cfg is None:
            self._cfg = datalad.cfg
        return self._cfg

    def __enter__(self):
        """Default implementation that does nothing in particular"""
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Default implementation that only calls ``.close()``"""
        self.close()
        # we have no desire to suppress exception, indicate standard
        # handling by not returning True
        return

    @contextmanager
    @abstractmethod
    def open(self, item: Any) -> Generator[IO | None, None, None]:
        """Get a file-like for an archive item

        Parameters
        ----------
        item:
          Any identifier for an archive item supported by a particular handler
        """
        raise NotImplementedError

    def close(self) -> None:
        """Default implementation for closing a archive handler

        This default implementation does nothing.
        """
        pass

    @abstractmethod
    def __contains__(self, item: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def __iter__(self) -> Generator[FileSystemItem, None, None]:
        raise NotImplementedError




from ..requests_auth import parse_www_authenticate


challenges = (
    # just challenge type
    ('Negotiate',
     [('negotiate', None)]),
    # challenge and just a token, tolerate any base64 padding
    ('Negotiate abcdef',
     [('negotiate', 'abcdef')]),
    ('Negotiate abcdef=',
     [('negotiate', 'abcdef=')]),
    ('Negotiate abcdef==',
     [('negotiate', 'abcdef==')]),
    # standard bearer
    ('Bearer realm=example.com',
     [('bearer', {'realm': 'example.com'})]),
    # standard digest
    ('Digest realm="example.com", qop="auth,auth-int", nonce="abcdef", '
     'opaque="ghijkl"',
     [('digest', {'realm': 'example.com', 'qop': 'auth,auth-int',
                  'nonce': 'abcdef', 'opaque': 'ghijkl'})]),
    # multi challenge
    ('Basic speCial="paf ram", realm="basIC", '
     'Bearer, '
     'Digest realm="http-auth@example.org", qop="auth, auth-int", '
     'algorithm=MD5',
     [('basic', {'special': 'paf ram', 'realm': 'basIC'}),
      ('bearer', None),
      ('digest', {'realm': "http-auth@example.org", 'qop': "auth, auth-int",
                  'algorithm': 'MD5'})]),
    # same challenge, multiple times, last one wins
    ('Basic realm="basIC", '
     'Basic realm="complex"',
     [('basic', {'realm': 'complex'})]),
)


def test_parse_www_authenticate():
    for hdr, targets in challenges:
        res = parse_www_authenticate(hdr)
        for ctype, props in targets:
            assert ctype in res
            assert res[ctype] == props



""" Patch ``datalad.distributed.ora_remote.ORARemote``

This patch replaces the class :class:`datalad.distributed.ora_remote.ORARemote`
with an updated version that should work properly on Linux, OSX, and Windows.

The main difference to the original code is that all path-operations are
performed on URL-paths. Those are represented by instances of `PurePosixPath`.
All subclasses of :class:`BaseIO`, i.e. :class:`LocalIO`, :class:`SSHRemoteIO`,
and :class:`HTTPRemoteIO`, are extended to contain the method
:meth:`url2transport_path`. This method converts an URL-path into the correct
path for the transport, i.e. the IO abstraction.

Before methods on a subclass of :class:`BaseIO` that require a path are called,
the generic URL-path is converted into the correct path for the IO-class by
calling :meth:`url2transport_path` on the respective IO-class.

The patch keeps changes to the necessary minimum. That means the source
is mostly identical to the original. Besides the changes described above,
more debug output was added.

NOTE: this patch only provides :class:`ORARemote`. The patches that add a
:meth:`url2transport_path`-method to :class:`LocalIO` and to :class:`HTTPRemoteIO`
are contained in module ``datalad_next.patches.add_method_url2localpath``. The
reason to keep them separate is that the patch from module
``datalad_next.patches.replace_create_sibling_ria`` require them as well.
For :class:`SSHRemoteIO` the method is included in the patch definition of
:class:`SSHRemoteIO`, which is contained in the module
``datalad_next.patches.replace_sshremoteio``.
"""
from __future__ import annotations

import re
import urllib.parse
from pathlib import (
    Path,
    PurePosixPath,
)
from shlex import quote as sh_quote
import logging

from datalad.config import anything2bool
from datalad.customremotes import (
    ProtocolError,
    SpecialRemote,
)
from datalad.distributed.ora_remote import (
    HTTPRemoteIO,
    LocalIO,
    RIARemoteError,
    SSHRemoteIO,
    _get_datalad_id,
    _get_gitcfg,
    handle_errors,
    NoLayoutVersion,
)
from datalad.support.annex_utils import _sanitize_key
from datalad.support.annexrepo import AnnexRepo
from datalad.customremotes.ria_utils import (
    get_layout_locations,
    UnknownLayoutVersion,
    verify_ria_url,
)
from datalad.utils import on_windows

from . import apply_patch


lgr = logging.getLogger('datalad.customremotes.ria_remote')

drive_letter_matcher = re.compile('^[A-Z]:')
slash_drive_letter_matcher = re.compile('^/[A-Z]:')

DEFAULT_BUFFER_SIZE = 65536


def canonify_url(url: str | None):
    """For file URLs on windows: put the drive letter into the path component"""
    if not on_windows or url is None:
        return url
    url_parts = urllib.parse.urlparse(url)
    if url_parts.scheme not in ('ria+file', 'file'):
        return url

    match = drive_letter_matcher.match(url_parts.netloc)
    if not match:
        return url

    return f'{url_parts.scheme}:///{match.string}{url_parts.path}'


def de_canonify_url(url: str | None):
    """For file URLs on windows: put the drive letter into the netloc component"""
    if not on_windows or url is None:
        return url
    url_parts = urllib.parse.urlparse(url)
    if url_parts.scheme not in ('ria+file', 'file'):
        return url

    match = slash_drive_letter_matcher.match(url_parts.path)
    if not match:
        return url

    return f'{url_parts.scheme}://{url_parts.path[1:3]}{url_parts.path[3:]}'


# `ORARemote` taken from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732:
class ORARemote(SpecialRemote):
    """This is the class of RIA remotes.
    """

    dataset_tree_version = '1'
    object_tree_version = '2'
    # TODO: Move known versions. Needed by creation routines as well.
    known_versions_objt = ['1', '2']
    known_versions_dst = ['1']

    @handle_errors
    def __init__(self, annex):
        super(ORARemote, self).__init__(annex)
        if hasattr(self, 'configs'):
            # introduced in annexremote 1.4.2 to support LISTCONFIGS
            self.configs['url'] = "RIA store to use"
            self.configs['push-url'] = "URL for pushing to the RIA store. " \
                                       "Optional."
            self.configs['archive-id'] = "Dataset ID (fallback: annex uuid. " \
                                         "Should be set automatically by " \
                                         "datalad"
        # the local repo
        self._repo = None
        self.gitdir = None
        self.name = None  # name of the special remote
        self.gitcfg_name = None  # name in respective git remote

        self.ria_store_url = None
        self.ria_store_pushurl = None
        # machine to SSH-log-in to access/store the data
        # subclass must set this
        self.storage_host = None
        self.storage_host_push = None
        # must be absolute, and POSIX (will be instance of PurePosixPath)
        # subclass must set this
        self.store_base_path = None
        self.store_base_path_push = None
        # by default we can read and write
        self.read_only = False
        self.force_write = None
        self.ignore_remote_config = None
        self.remote_log_enabled = None
        self.remote_dataset_tree_version = None
        self.remote_object_tree_version = None

        # for caching the remote's layout locations:
        self.remote_git_dir = None
        self.remote_archive_dir = None
        self.remote_obj_dir = None
        # lazy IO:
        self._io = None
        self._push_io = None

        # cache obj_locations:
        self._last_archive_path = None
        self._last_keypath = (None, None)

        # SSH "streaming" buffer
        self.buffer_size = DEFAULT_BUFFER_SIZE

    # PATCH: add a helper to assert the type of a path.
    @staticmethod
    def _assert_pure_posix_path(path):
        assert isinstance(path, PurePosixPath)

    # PATCH: add a close function to ensure that all IO-abstraction objects are
    # closed.
    def close(self):
        if self._io:
            self._io.close()
            self._io = None
        if self._push_io:
            self._push_io.close()
            self._push_io = None

    def verify_store(self):
        """Check whether the store exists and reports a layout version we
        know

        The layout of the store is recorded in base_path/ria-layout-version.
        If the version found on the remote end isn't supported and `force-write`
        isn't configured, sets the remote to read-only operation.
        """
        # THE PATCH: assert path type and perform operation on abstract path
        self._assert_pure_posix_path(self.store_base_path)
        dataset_tree_version_file = self.store_base_path / 'ria-layout-version'

        # check dataset tree version
        try:
            self.remote_dataset_tree_version = \
                self._get_version_config(dataset_tree_version_file)
        except Exception as exc:
            raise RIARemoteError("RIA store unavailable.") from exc
        if self.remote_dataset_tree_version not in self.known_versions_dst:
            # Note: In later versions, condition might change in order to
            # deal with older versions.
            raise UnknownLayoutVersion(f"RIA store layout version unknown: "
                                       f"{self.remote_dataset_tree_version}")

    def verify_ds_in_store(self):
        """Check whether the dataset exists in store and reports a layout
        version we know

        The layout is recorded in
        'dataset_somewhere_beneath_base_path/ria-layout-version.'
        If the version found on the remote end isn't supported and `force-write`
        isn't configured, sets the remote to read-only operation.
        """

        object_tree_version_file = self.remote_git_dir / 'ria-layout-version'

        # check (annex) object tree version
        try:
            self.remote_object_tree_version =\
                self._get_version_config(object_tree_version_file)
        except Exception as e:
            raise RIARemoteError("Dataset unavailable from RIA store.")
        if self.remote_object_tree_version not in self.known_versions_objt:
            raise UnknownLayoutVersion(f"RIA dataset layout version unknown: "
                                       f"{self.remote_object_tree_version}")

    def _load_local_cfg(self):

        # this will work, even when this is not a bare repo
        # but it is not capable of reading out dataset/branch config
        self._repo = AnnexRepo(self.gitdir)

        cfg_map = {"ora-force-write": "force_write",
                   "ora-ignore-ria-config": "ignore_remote_config",
                   "ora-buffer-size": "buffer_size",
                   "ora-url": "ria_store_url",
                   "ora-push-url": "ria_store_pushurl"
                   }

        # in initremote we may not have a reliable name of the git remote config
        # yet. Go with the default.
        gitcfg_name = self.gitcfg_name or self.name
        if gitcfg_name:
            for cfg, att in cfg_map.items():
                value = self._repo.config.get(f"remote.{gitcfg_name}.{cfg}")
                if value is not None:
                    self.__setattr__(cfg_map[cfg], value)
                    if cfg == "ora-url":
                        self.ria_store_url_source = 'local'
                    elif cfg == "ora-push-url":
                        self.ria_store_pushurl_source = 'local'
            if self.buffer_size:
                try:
                    self.buffer_size = int(self.buffer_size)
                except ValueError:
                    self.message(f"Invalid value of config "
                                 f"'remote.{gitcfg_name}."
                                 f"ora-buffer-size': {self.buffer_size}")
                    self.buffer_size = DEFAULT_BUFFER_SIZE

        if self.name:
            # Consider deprecated configs if there's no value yet
            if self.force_write is None:
                self.force_write = self._repo.config.get(
                    f'annex.ora-remote.{self.name}.force-write')
                if self.force_write:
                    self.message("WARNING: config "
                                 "'annex.ora-remote.{}.force-write' is "
                                 "deprecated. Use 'remote.{}.ora-force-write' "
                                 "instead.".format(self.name, self.gitcfg_name))
                    try:
                        self.force_write = anything2bool(self.force_write)
                    except TypeError:
                        raise RIARemoteError("Invalid value of config "
                                             "'annex.ora-remote.{}.force-write'"
                                             ": {}".format(self.name,
                                                           self.force_write))

            if self.ignore_remote_config is None:
                self.ignore_remote_config = self._repo.config.get(
                    f"annex.ora-remote.{self.name}.ignore-remote-config")
                if self.ignore_remote_config:
                    self.message("WARNING: config "
                                 "'annex.ora-remote.{}.ignore-remote-config' is"
                                 " deprecated. Use "
                                 "'remote.{}.ora-ignore-ria-config' instead."
                                 "".format(self.name, self.gitcfg_name))
                    try:
                        self.ignore_remote_config = \
                            anything2bool(self.ignore_remote_config)
                    except TypeError:
                        raise RIARemoteError(
                            "Invalid value of config "
                            "'annex.ora-remote.{}.ignore-remote-config': {}"
                            "".format(self.name, self.ignore_remote_config))

    def _load_committed_cfg(self, fail_noid=True):

        # which repo are we talking about
        self.gitdir = self.annex.getgitdir()

        # go look for an ID
        self.archive_id = self.annex.getconfig('archive-id')
        if fail_noid and not self.archive_id:
            # TODO: Message! "archive ID" is confusing. dl-id or annex-uuid
            raise RIARemoteError(
                "No archive ID configured. This should not happen.")

        # what is our uuid?
        self.uuid = self.annex.getuuid()

        # RIA store URL(s)
        self.ria_store_url = self.annex.getconfig('url')
        if self.ria_store_url:
            self.ria_store_url_source = 'annex'
        self.ria_store_pushurl = self.annex.getconfig('push-url')
        if self.ria_store_pushurl:
            self.ria_store_pushurl_source = 'annex'

        # TODO: This should prob. not be done! Would only have an effect if
        #       force-write was committed annex-special-remote-config and this
        #       is likely a bad idea.
        self.force_write = self.annex.getconfig('force-write')
        if self.force_write == "":
            self.force_write = None

        # Get the special remote name
        # TODO: Make 'name' a property of `SpecialRemote`;
        #       Same for `gitcfg_name`, `_repo`?
        self.name = self.annex.getconfig('name')
        if not self.name:
            self.name = self.annex.getconfig('sameas-name')
        if not self.name:
            # TODO: Do we need to crash? Not necessarily, I think. We could
            #       still find configs and if not - might work out.
            raise RIARemoteError(
                "Cannot determine special remote name, got: {}".format(
                    repr(self.name)))
        # Get the name of the remote entry in .git/config.
        # Note, that this by default is the same as the stored name of the
        # special remote, but can be different (for example after
        # git-remote-rename). The actual connection is the uuid of the special
        # remote, not the name.
        try:
            self.gitcfg_name = self.annex.getgitremotename()
        except (ProtocolError, AttributeError):
            # GETGITREMOTENAME not supported by annex version or by annexremote
            # version.
            # Lets try to find ourselves: Find remote with matching annex uuid
            response = _get_gitcfg(self.gitdir,
                                   r"^remote\..*\.annex-uuid",
                                   regex=True)
            response = response.splitlines() if response else []
            candidates = set()
            for line in response:
                k, v = line.split()
                if v == self.annex.getuuid():  # TODO: Where else? self.uuid?
                    candidates.add(''.join(k.split('.')[1:-1]))
            num_candidates = len(candidates)
            if num_candidates == 1:
                self.gitcfg_name = candidates.pop()
            elif num_candidates > 1:
                self.message("Found multiple used remote names in git "
                             "config: %s" % str(candidates))
                # try same name:
                if self.name in candidates:
                    self.gitcfg_name = self.name
                    self.message("Choose '%s'" % self.name)
                else:
                    self.gitcfg_name = None
                    self.message("Ignore git config")
            else:
                # No entry found.
                # Possible if we are in "initremote".
                self.gitcfg_name = None

    def _load_cfg(self, gitdir, name):
        # Whether or not to force writing to the remote. Currently used to
        # overrule write protection due to layout version mismatch.
        self.force_write = self._repo.config.get(
            f'annex.ora-remote.{name}.force-write')

        # whether to ignore config flags set at the remote end
        self.ignore_remote_config = \
            self._repo.config.get(
                f'annex.ora-remote.{name}.ignore-remote-config')

        # buffer size for reading files over HTTP and SSH
        self.buffer_size = self._repo.config.get(
            f"remote.{name}.ora-buffer-size")

        if self.buffer_size:
            self.buffer_size = int(self.buffer_size)

    def _verify_config(self, fail_noid=True):
        # try loading all needed info from (git) config

        # first load committed config
        self._load_committed_cfg(fail_noid=fail_noid)
        # now local configs (possible overwrite of committed)
        self._load_local_cfg()
        # PATCH: use canonified URLs
        self.ria_store_url = canonify_url(self.ria_store_url)
        self.ria_store_pushurl = canonify_url(self.ria_store_pushurl)

        # get URL rewriting config
        url_cfgs = {k: v for k, v in self._repo.config.items()
                    if k.startswith('url.')}

        if self.ria_store_url:
            self.storage_host, self.store_base_path, self.ria_store_url = \
                verify_ria_url(self.ria_store_url, url_cfgs)

        else:
            # There's one exception to the precedence of local configs:
            # Age-old "ssh-host" + "base-path" configs are only considered,
            # if there was no RIA URL (local or committed). However, issue
            # deprecation warning, if that situation is encountered:
            host = None
            path = None

            if self.name:
                host = self._repo.config.get(
                    f'annex.ora-remote.{self.name}.ssh-host') or \
                       self.annex.getconfig('ssh-host')
                # Note: Special value '0' is replaced by None only after checking
                # the repository's annex config. This is to uniformly handle '0' and
                # None later on, but let a user's config '0' overrule what's
                # stored by git-annex.
                self.storage_host = None if host == '0' else host
                path = self._repo.config.get(
                    f'annex.ora-remote.{self.name}.base-path') or \
                       self.annex.getconfig('base-path')
                self.store_base_path = path.strip() if path else path

            if path or host:
                self.message("WARNING: base-path + ssh-host configs are "
                             "deprecated and won't be considered in the future."
                             " Use 'git annex enableremote {} "
                             "url=<RIA-URL-TO-STORE>' to store a ria+<scheme>:"
                             "//... URL in the special remote's config."
                             "".format(self.name),
                             type='info')


        if not self.store_base_path:
            raise RIARemoteError(
                "No base path configured for RIA store. Specify a proper "
                "ria+<scheme>://... URL.")

        # the base path is ultimately derived from a URL, always treat as POSIX
        self.store_base_path = PurePosixPath(self.store_base_path)
        if not self.store_base_path.is_absolute():
            raise RIARemoteError(
                'Non-absolute RIA store base path configuration: %s'
                '' % str(self.store_base_path))

        if self.ria_store_pushurl:
            if self.ria_store_pushurl.startswith("ria+http"):
                raise RIARemoteError("Invalid push-url: {}. Pushing over HTTP "
                                     "not implemented."
                                     "".format(self.ria_store_pushurl))
            self.storage_host_push, \
            self.store_base_path_push, \
            self.ria_store_pushurl = \
                verify_ria_url(self.ria_store_pushurl, url_cfgs)
            self.store_base_path_push = PurePosixPath(self.store_base_path_push)

    def _get_version_config(self, path):
        """ Get version and config flags from RIA store's layout file
        """

        if self.ria_store_url:
            # construct path to ria_layout_version file for reporting
            # PATCH: use abstract path
            local_store_base_path = self.store_base_path
            target_ri = (
                self.ria_store_url[4:]
                + "/"
                + path.relative_to(local_store_base_path).as_posix()
            )
        elif self.storage_host:
            target_ri = "ssh://{}{}".format(self.storage_host, path.as_posix())
        else:
            target_ri = path.as_uri()

        try:
            # PATCH: convert abstract path to io-specific concrete path
            file_content = self.io.read_file(
                self.io.url2transport_path(path)
            ).strip().split('|')

        # Note, that we enhance the reporting here, as the IO classes don't
        # uniformly operate on that kind of RI (which is more informative
        # as it includes the store base address including the access
        # method).
        except FileNotFoundError as exc:
            raise NoLayoutVersion(
                f"{target_ri} not found, "
                f"self.ria_store_url: {self.ria_store_url}, "
                f"self.store_base_path: {self.store_base_path}, "
                f"self.store_base_path_push: {self.store_base_path_push}, "
                f"path: {type(path)} {path}") from exc
        except PermissionError as exc:
            raise PermissionError(f"Permission denied: {target_ri}") from exc
        except Exception as exc:
            raise RuntimeError(f"Failed to access {target_ri}") from exc

        if not (1 <= len(file_content) <= 2):
            self.message("invalid version file {}".format(path),
                         type='info')
            return None

        remote_version = file_content[0]
        remote_config_flags = file_content[1] \
            if len(file_content) == 2 else None
        if not self.ignore_remote_config and remote_config_flags:
            # Note: 'or', since config flags can come from toplevel
            #       (dataset-tree-root) as well as from dataset-level.
            #       toplevel is supposed flag the entire tree.
            self.remote_log_enabled = self.remote_log_enabled or \
                                      'l' in remote_config_flags

        return remote_version

    def get_store(self):
        """checks the remote end for an existing store and dataset

        Furthermore reads and stores version and config flags, layout
        locations, etc.
        If this doesn't raise, the remote end should be fine to work with.
        """
        # make sure the base path is a platform path when doing local IO
        # the incoming Path object is a PurePosixPath
        # XXX this else branch is wrong: Incoming is PurePosixPath
        # but it is subsequently assumed to be a platform path, by
        # get_layout_locations() etc. Hence it must be converted
        # to match the *remote* platform, not the local client

        # cache remote layout directories
        # PATCH: use the abstract `self.store_base_path` to calculate RIA-store
        # directory paths.
        self.remote_git_dir, self.remote_archive_dir, self.remote_obj_dir = \
            self.get_layout_locations(self.store_base_path, self.archive_id)

        read_only_msg = "Treating remote as read-only in order to " \
                        "prevent damage by putting things into an unknown " \
                        "version of the target layout. You can overrule this " \
                        "by setting 'annex.ora-remote.<name>.force-write=true'."
        try:
            self.verify_store()
        except UnknownLayoutVersion:
            reason = "Remote dataset tree reports version {}. Supported " \
                     "versions are: {}. Consider upgrading datalad or " \
                     "fix the 'ria-layout-version' file at the RIA store's " \
                     "root. ".format(self.remote_dataset_tree_version,
                                     self.known_versions_dst)
            self._set_read_only(reason + read_only_msg)
        except NoLayoutVersion:
            reason = "Remote doesn't report any dataset tree version. " \
                     "Consider upgrading datalad or add a fitting " \
                     "'ria-layout-version' file at the RIA store's " \
                     "root."
            self._set_read_only(reason + read_only_msg)

        try:
            self.verify_ds_in_store()
        except UnknownLayoutVersion:
            reason = "Remote object tree reports version {}. Supported" \
                     "versions are {}. Consider upgrading datalad or " \
                     "fix the 'ria-layout-version' file at the remote " \
                     "dataset root. " \
                     "".format(self.remote_object_tree_version,
                               self.known_versions_objt)
            self._set_read_only(reason + read_only_msg)
        except NoLayoutVersion:
            reason = "Remote doesn't report any object tree version. " \
                     "Consider upgrading datalad or add a fitting " \
                     "'ria-layout-version' file at the remote " \
                     "dataset root. "
            self._set_read_only(reason + read_only_msg)

    @handle_errors
    def initremote(self):
        self._verify_config(fail_noid=False)
        if not self.archive_id:
            self.archive_id = _get_datalad_id(self.gitdir)
            if not self.archive_id:
                # fall back on the UUID for the annex remote
                self.archive_id = self.annex.getuuid()

        self.get_store()

        self.annex.setconfig('archive-id', self.archive_id)
        # Make sure, we store the potentially rewritten URL. But only, if the
        # source was annex as opposed to a local config.
        if self.ria_store_url and self.ria_store_url_source == 'annex':
            self.annex.setconfig('url', self.ria_store_url)
        if self.ria_store_pushurl and self.ria_store_pushurl_source == 'annex':
            self.annex.setconfig('push-url', self.ria_store_pushurl)

    def _local_io(self):
        """Are we doing local operations?"""
        # let's not make this decision dependent on the existence
        # of a directory the matches the name of the configured
        # store tree base dir. Such a match could be pure
        # coincidence. Instead, let's do remote whenever there
        # is a remote host configured
        #return self.store_base_path.is_dir()

        # TODO: Isn't that wrong with HTTP anyway?
        #       + just isinstance(LocalIO)?
        # XXX isinstance(LocalIO) would not work, this method is used
        # before LocalIO is instantiated
        return not self.storage_host

    def _set_read_only(self, msg):

        if not self.force_write:
            self.read_only = True
            self.message(msg, type='info')
        else:
            self.message("Was instructed to force write", type='info')

    def _ensure_writeable(self):
        if self.read_only:
            raise RIARemoteError("Remote is treated as read-only. "
                                 "Set 'ora-remote.<name>.force-write=true' to "
                                 "overrule this.")
        if isinstance(self.push_io, HTTPRemoteIO):
            raise RIARemoteError("Write access via HTTP not implemented")

    @property
    def io(self):
        if not self._io:
            if self._local_io():
                self._io = LocalIO()
            elif self.ria_store_url.startswith("ria+http"):
                # TODO: That construction of "http(s)://host/" should probably
                #       be moved, so that we get that when we determine
                #       self.storage_host. In other words: Get the parsed URL
                #       instead and let HTTPRemoteIO + SSHRemoteIO deal with it
                #       uniformly. Also: Don't forget about a possible port.

                url_parts = self.ria_store_url[4:].split('/')
                # we expect parts: ("http(s):", "", host:port, path)
                self._io = HTTPRemoteIO(
                    url_parts[0] + "//" + url_parts[2],
                    self.buffer_size
                )
            elif self.storage_host:
                self._io = SSHRemoteIO(self.storage_host, self.buffer_size)
                from atexit import register
                register(self._io.close)
            else:
                raise RIARemoteError(
                    "Local object tree base path does not exist, and no SSH"
                    "host configuration found.")
        return self._io

    @property
    def push_io(self):
        # Instance of an IOBase subclass for execution based on configured
        # 'push-url' if such exists. Otherwise identical to `self.io`.
        # Note, that once we discover we need to use the push-url (that is on
        # TRANSFER_STORE and REMOVE), we should switch all operations to that IO
        # instance instead of using different connections for read and write
        # operations. Ultimately this is due to the design of annex' special
        # remote protocol - we don't know which annex command is running and
        # therefore we don't know whether to use fetch or push URL during
        # PREPARE.

        if not self._push_io:
            if self.ria_store_pushurl:
                self.message("switching ORA to push-url")
                # Not-implemented-push-HTTP is ruled out already when reading
                # push-url, so either local or SSH:
                if not self.storage_host_push:
                    # local operation
                    self._push_io = LocalIO()
                else:
                    self._push_io = SSHRemoteIO(self.storage_host_push,
                                                self.buffer_size)

                # We have a new instance. Kill the existing one and replace.
                from atexit import register, unregister
                if hasattr(self.io, 'close'):
                    unregister(self.io.close)
                    self.io.close()

                # XXX now also READ IO is done with the write IO
                # this explicitly ignores the remote config
                # that distinguishes READ from WRITE with different
                # methods
                self._io = self._push_io
                if hasattr(self.io, 'close'):
                    register(self.io.close)

                self.storage_host = self.storage_host_push
                self.store_base_path = self.store_base_path_push

                # delete/update cached locations:
                self._last_archive_path = None
                self._last_keypath = (None, None)

                self.remote_git_dir, \
                self.remote_archive_dir, \
                self.remote_obj_dir = \
                    self.get_layout_locations(
                        # PATCH: use abstract path to calculate RIA-store dirs
                        self.store_base_path,
                        self.archive_id
                    )

            else:
                # no push-url: use existing IO
                self._push_io = self._io

        return self._push_io

    @handle_errors
    def prepare(self):

        gitdir = self.annex.getgitdir()
        self._repo = AnnexRepo(gitdir)
        self._verify_config()

        self.get_store()

        # report active special remote configuration/status
        self.info = {
            'store_base_path': str(self.store_base_path),
            'storage_host': 'local'
            if self._local_io() else self.storage_host,
        }

        # TODO: following prob. needs hasattr instead:
        if not isinstance(self.io, HTTPRemoteIO):
            self.info['7z'] = ("not " if not self.io.get_7z() else "") + \
                              "available"

    @handle_errors
    def transfer_store(self, key, filename):
        self._ensure_writeable()

        # we need a file-system compatible name for the key
        key = _sanitize_key(key)

        dsobj_dir, archive_path, key_path = self._get_obj_location(key)
        key_path = dsobj_dir / key_path

        # PATCH: convert abstract path `key_path` to io-specific concrete path
        # and use that.
        transport_key_path = self.push_io.url2transport_path(key_path)
        if self.push_io.exists(transport_key_path):
            # if the key is here, we trust that the content is in sync
            # with the key
            return

        self.push_io.mkdir(transport_key_path.parent)

        # We need to copy to a temp location to let checkpresent fail while the
        # transfer is still in progress and furthermore not interfere with
        # administrative tasks in annex/objects.
        # In addition include uuid, to not interfere with parallel uploads from
        # different clones.
        transfer_dir = \
            self.remote_git_dir / "ora-remote-{}".format(self._repo.uuid) / "transfer"
        # PATCH: convert abstract path `transfer_dir` to io-specific concrete
        # path and use that
        transport_transfer_dir = self.push_io.url2transport_path(transfer_dir)
        self.push_io.mkdir(transport_transfer_dir)

        tmp_path = transfer_dir / key
        # PATCH: convert abstract path `transport_tmp_path` to io-specific
        # concrete path and use that
        transport_tmp_path = self.push_io.url2transport_path(tmp_path)
        try:
            self.push_io.put(filename, transport_tmp_path, self.annex.progress)
            # copy done, atomic rename to actual target
            self.push_io.rename(transport_tmp_path, transport_key_path)
        except Exception as e:
            # whatever went wrong, we don't want to leave the transfer location
            # blocked
            self.push_io.remove(transport_tmp_path)
            raise e

    @handle_errors
    def transfer_retrieve(self, key, filename):
        # we need a file-system compatible name for the key
        key = _sanitize_key(key)

        dsobj_dir, archive_path, key_path = self._get_obj_location(key)
        abs_key_path = dsobj_dir / key_path
        # PATCH: convert abstract path `abs_key_path` to io-specific
        # concrete path and use that
        transport_abs_key_path = self.io.url2transport_path(abs_key_path)

        # sadly we have no idea what type of source gave checkpresent->true
        # we can either repeat the checks, or just make two opportunistic
        # attempts (at most)
        try:
            self.io.get(transport_abs_key_path, filename, self.annex.progress)
        except Exception as e1:
            if isinstance(self.io, HTTPRemoteIO):
                # no client-side archive access over HTTP
                # Note: This is intentional, as it would mean one additional
                # request per key. However, server response to the GET can
                # consider archives on their end.
                raise
            # catch anything and keep it around for a potential re-raise
            try:
                # PATCH: convert abstract path `archive_path` to io-specific
                # concrete path and use that
                transport_archive_path = self.io.url2transport_path(
                    archive_path
                )
                self.io.get_from_archive(
                    transport_archive_path,
                    key_path,
                    filename,
                    self.annex.progress
                )
            except Exception as e2:
                # TODO properly report the causes
                raise RIARemoteError('Failed to obtain key: {}'
                                     ''.format([str(e1), str(e2)]))

    @handle_errors
    def checkpresent(self, key):
        # we need a file-system compatible name for the key
        key = _sanitize_key(key)

        dsobj_dir, archive_path, key_path = self._get_obj_location(key)
        abs_key_path = dsobj_dir / key_path
        # PATCH: convert abstract path `abs_key_path` to io-specific concrete
        # path and use that
        transport_abs_key_path = self.io.url2transport_path(abs_key_path)
        if self.io.exists(transport_abs_key_path):
            # we have an actual file for this key
            return True
        if isinstance(self.io, HTTPRemoteIO):
            # no client-side archive access over HTTP
            return False
        # do not make a careful check whether an archive exists, because at
        # present this requires an additional SSH call for remote operations
        # which may be rather slow. Instead just try to run 7z on it and let
        # it fail if no archive is around
        # TODO honor future 'archive-mode' flag
        # PATCH: convert abstract path `archive_path` to io-specific concrete
        # path and use that
        transport_archive_path = self.io.url2transport_path(archive_path)
        return self.io.in_archive(transport_archive_path, key_path)

    @handle_errors
    def remove(self, key):
        # we need a file-system compatible name for the key
        key = _sanitize_key(key)

        self._ensure_writeable()

        dsobj_dir, archive_path, key_path = self._get_obj_location(key)
        key_path = dsobj_dir / key_path
        # PATCH: convert abstract path `key_path` to io-specific concrete path
        # and use that
        transport_key_path = self.push_io.url2transport_path(key_path)
        if self.push_io.exists(transport_key_path):
            self.push_io.remove(transport_key_path)
        key_dir = key_path
        # remove at most two levels of empty directories
        for level in range(2):
            key_dir = key_dir.parent
            try:
                # PATCH: convert abstract path `key_dir` to io-specific concrete
                # path and use that
                transport_key_dir = self.push_io.url2transport_path(key_dir)
                self.push_io.remove_dir(transport_key_dir)
            except Exception:
                break

    @handle_errors
    def getcost(self):
        # 100 is cheap, 200 is expensive (all relative to Config/Cost.hs)
        # 100/200 are the defaults for local and remote operations in
        # git-annex
        # if we have the object tree locally, operations are cheap (100)
        # otherwise expensive (200)
        return '100' if self._local_io() else '200'

    @handle_errors
    def whereis(self, key):
        # we need a file-system compatible name for the key
        key = _sanitize_key(key)

        dsobj_dir, archive_path, key_path = self._get_obj_location(key)
        if isinstance(self.io, HTTPRemoteIO):
            # display the URL for a request
            # TODO: method of HTTPRemoteIO
            # in case of a HTTP remote (unchecked for others), storage_host
            # is not just a host, but a full URL without a path
            return f'{self.storage_host}{dsobj_dir}/{key_path}'

        return str(dsobj_dir / key_path) if self._local_io() \
            else '{}: {}:{}'.format(
                self.storage_host,
                self.remote_git_dir,
                sh_quote(str(key_path)),
        )

    @staticmethod
    def get_layout_locations(base_path, dsid):
        # PATCH: type of `base_path` is `PurePosixPath`
        ORARemote._assert_pure_posix_path(base_path)
        return get_layout_locations(1, base_path, dsid)

    def _get_obj_location(self, key):
        # Notes: - Changes to this method may require an update of
        #          ORARemote._layout_version
        #        - archive_path is always the same ATM. However, it might depend
        #          on `key` in the future. Therefore build the actual filename
        #          for the archive herein as opposed to `get_layout_locations`.

        if not self._last_archive_path:
            # PATCH: type of `base_path` is `PurePosixPath`
            self._assert_pure_posix_path(self.remote_archive_dir)
            self._last_archive_path = self.remote_archive_dir / 'archive.7z'

        if self._last_keypath[0] != key:
            if self.remote_object_tree_version == '1':
                # PATCH: dir-hashes are always in platform format. We convert it
                # to a platform-specific `Path` and then to `PurePosixPath`.
                key_dir = PurePosixPath(Path(self.annex.dirhash_lower(key)))

            # If we didn't recognize the remote layout version, we set to
            # read-only and promised to at least try and read according to our
            # current version. So, treat that case as if remote version was our
            # (client's) version.
            else:
                # PATCH: dir-hashes are always in platform format. We convert it
                # to a platform-specific `Path` and then to `PurePosixPath`.
                key_dir = PurePosixPath(Path(self.annex.dirhash(key)))
            # double 'key' is not a mistake, but needed to achieve the exact
            # same layout as the annex/objects tree
            # PATCH: use the abstract `key_dir` path
            self._last_keypath = (key, key_dir / key / key)

        self._assert_pure_posix_path(self.remote_obj_dir)
        return self.remote_obj_dir, self._last_archive_path, \
            self._last_keypath[1]


apply_patch(
    'datalad.distributed.ora_remote',
    None,
    'ORARemote',
    ORARemote,
)



import warnings
warnings.warn(
    "datalad_next.utils.credman was replaced by datalad_next.credman in "
    "datalad_next 1.0. This transition helper module will be removed in "
    "datalad_next 2.0.",
    DeprecationWarning,
)

from datalad_next.credman.manager import (
    CredentialManager,
    verify_property_names,
)



import pytest

from ..git import (
    CommandError,
    call_git,
    call_git_lines,
    call_git_oneline,
    call_git_success,
    iter_git_subproc,
)


def test_call_git():
    # smoke test
    call_git(['--version'])
    # raises properly
    with pytest.raises(CommandError):
        call_git(['notacommand'])


def test_call_git_success():
    assert call_git_success(['--version'])
    assert not call_git_success(['notacommand'])


def test_call_git_lines():
    lines = call_git_lines(['--version'])
    assert len(lines) == 1
    assert lines[0].startswith('git version')
    # check that we can force Git into LC_ALL mode.
    # this test is only meaningful on systems that
    # run with some other locale
    call_git_lines(['-h'])[0].casefold().startswith('usage')


def test_call_git_oneline():
    line = call_git_oneline(['--version'])
    assert line.startswith('git version')
    with pytest.raises(AssertionError):
        # TODO may not yield multiple lines on all systems
        call_git_oneline(['config', '-l'])


def test_iter_git_subproc():
    # just a smoke test that 'git' gets prepended
    with iter_git_subproc(['--version']) as g:
        assert list(g)



"""Report on the content of a Git tree-ish

The main functionality is provided by the :func:`iter_gittree()` function.
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from functools import cached_property
import logging
from pathlib import (
    Path,
    PurePosixPath,
)
from typing import (
    Iterator,
    Generator,
)
from datasalad.itertools import (
    decode_bytes,
    itemize,
)

from datalad_next.runners import iter_git_subproc

from .utils import PathBasedItem


lgr = logging.getLogger('datalad.ext.next.iter_collections.gittree')


# TODO Could be `StrEnum`, came with PY3.11
class GitTreeItemType(Enum):
    """Enumeration of item types of Git trees
    """
    file = 'file'
    executablefile = 'executablefile'
    symlink = 'symlink'
    directory = 'directory'
    submodule = 'submodule'


@dataclass
class GitTreeItem(PathBasedItem):
    """``PathBasedItem`` with a relative path as a name (in POSIX conventions)
    """
    name: str
    # gitsha is not the sha1 of the file content, but the output
    # of `git hash-object` which does something like
    # `printf "blob $(wc -c < "$file_name")\0$(cat "$file_name")" | sha1sum`
    gitsha: str | None = None
    gittype: GitTreeItemType | None = None

    @cached_property
    def path(self) -> PurePosixPath:
        """Returns the item name as a ``PurePosixPath`` instance"""
        return PurePosixPath(self.name)


_mode_type_map = {
    '100644': GitTreeItemType.file,
    '100755': GitTreeItemType.executablefile,
    '040000': GitTreeItemType.directory,
    '120000': GitTreeItemType.symlink,
    '160000': GitTreeItemType.submodule,
}


def iter_gittree(
    path: Path,
    treeish: str,
    *,
    recursive: str = 'repository',
) -> Generator[GitTreeItem, None, None]:
    """Uses ``git ls-tree`` to report on a tree in a Git repository

    Parameters
    ----------
    path: Path
      Path of a directory in a Git repository to report on. This directory
      need not be the root directory of the repository, but must be part of
      the repository. If the directory is not the root directory of a
      non-bare repository, the iterator is constrained to items underneath
      that directory.
    recursive: {'repository', 'no'}, optional
      Behavior for recursion into subtrees. By default (``repository``),
      all tree within the repository underneath ``path``) are reported,
      but not tree within submodules. If ``no``, only direct children
      are reported on.

    Yields
    ------
    :class:`GitTreeItem`
      The ``name`` attribute of an item is a ``str`` with the corresponding
      (relative) path, as reported by Git (in POSIX conventions).
    """
    # we force-convert to Path to give us the piece of mind we want.
    # The docs already ask for that, but it is easy to
    # forget/ignore and leads to non-obvious errors. Running this once is
    # a cheap safety net
    path = Path(path)

    # although it would be easy to also query the object size, we do not
    # do so, because it has a substantial runtime impact. It is unclear
    # what the main factor for the slowdown is, but in test cases I can
    # see 10x slower
    #lstree_args = ['--long']
    # we do not go for a custom format that would allow for a single split
    # by tab, because if we do, Git starts quoting paths with special
    # characters (like tab) again
    #lstree_args = ['--format=%(objectmode)%x09%(objectname)%x09%(path)']
    lstree_args = []
    if recursive == 'repository':
        lstree_args.append('-r')

    for line in _git_ls_tree(path, treeish, *lstree_args):
        yield _get_tree_item(line)


def _get_tree_item(spec: str) -> GitTreeItem:
    props, path = spec.split('\t', maxsplit=1)
    # 0::2 gets the first and third (last) item, effectively skippping the
    # type name (blob/tree etc.), we have the mode lookup for that, which
    # provides more detail
    mode, sha = props.split(' ')[0::2]
    return GitTreeItem(
        name=path,
        gitsha=sha,
        gittype=_mode_type_map[mode],
    )


def _git_ls_tree(path: Path, *args) -> Iterator[str]:
    with iter_git_subproc(
            [
                'ls-tree',
                # we rely on zero-byte splitting below
                '-z',
                # otherwise take whatever is coming in
                *args,
            ],
            cwd=path,
    ) as r:
        yield from itemize(
            decode_bytes(r),
            sep='\0',
            keep_ends=False,
        )



"""Report on the content of TAR archives

The main functionality is provided by the :func:`iter_tar()` function.
"""

from __future__ import annotations

from dataclasses import dataclass
from functools import cached_property
from pathlib import (
    Path,
    PurePosixPath,
)
import tarfile
from typing import (
    TYPE_CHECKING,
    Generator,
)

from .utils import (
    FileSystemItem,
    FileSystemItemType,
)


@dataclass  # sadly PY3.10+ only (kw_only=True)
class TarfileItem(FileSystemItem):
    name: str
    """TAR uses POSIX paths as item identifiers. Not all POSIX paths can
    be represented on all (non-POSIX) file systems, therefore the item
    name is represented in POSIX form, instead of in platform conventions.
    """
    link_target: str | None = None
    """Just as for ``name``, a link target is also reported in POSIX
    format."""

    @cached_property
    def path(self) -> PurePosixPath:
        """Returns the item name as a ``PurePosixPath`` instance"""
        return PurePosixPath(self.name)

    @cached_property
    def link_target_path(self) -> PurePosixPath | None:
        """Returns the link_target as a ``PurePosixPath`` instance"""
        return PurePosixPath(self.link_target) \
            if self.link_target is not None else None


def iter_tar(
    path: Path,
    *,
    fp: bool = False,
) -> Generator[TarfileItem, None, None]:
    """Uses the standard library ``tarfile`` module to report on TAR archives

    A TAR archive can represent more or less the full bandwidth of file system
    properties, therefore reporting on archive members is implemented
    similar to :func:`~datalad_next.iter_collections.directory.iter_dir()`.
    The iterator produces an :class:`TarfileItem` instance with standard
    information on file system elements, such as ``size``, or ``mtime``.

    Parameters
    ----------
    path: Path
      Path of the TAR archive to report content for (iterate over).
    fp: bool, optional
      If ``True``, each file-type item includes a file-like object
      to access the file's content. This file handle will be closed
      automatically when the next item is yielded or the function
      returns.

    Yields
    ------
    :class:`TarfileItem`
      The ``name`` attribute of an item is a ``str`` with the corresponding
      archive member name (in POSIX conventions).
    """
    with tarfile.open(path, 'r') as tar:
        for member in tar:
            # reduce the complexity of tar member types to the desired
            # level (i.e. disregard the diversity of special files and
            # block devices)
            mtype = FileSystemItemType.file if member.isreg() \
                else FileSystemItemType.directory if member.isdir() \
                else FileSystemItemType.symlink if member.issym() \
                else FileSystemItemType.hardlink if member.islnk() \
                else FileSystemItemType.specialfile
            item = TarfileItem(
                name=member.name,
                type=mtype,
                size=member.size,
                mode=member.mode,
                mtime=member.mtime,
                uid=member.uid,
                gid=member.gid,
                link_target=member.linkname
                if member.linkname else None,
            )
            if fp and mtype in (
                    FileSystemItemType.file, FileSystemItemType.hardlink):
                mf = tar.extractfile(member)
                if TYPE_CHECKING:
                    assert mf
                with mf as fileobj:
                    item.fp = fileobj
                    yield item
            else:
                yield item



from __future__ import annotations

from dataclasses import dataclass
from pathlib import (
    Path,
    PurePosixPath,
)
from typing import Generator

import pytest

from datalad_next.iter_collections import FileSystemItemType
from datalad_next.tests.marker import skipif_no_network

from ..tarfile import TarArchiveOperations


@dataclass
class _TestArchive:
    path: Path
    item_count: int
    content: bytes
    target_hash: dict[str, str]


@pytest.fixture(scope='session')
def structured_sample_tar_xz(
    sample_tar_xz
) -> Generator[_TestArchive, None, None]:
    yield _TestArchive(
        path=sample_tar_xz,
        item_count=6,
        content=b'123\n',
        target_hash={
            'SHA1': 'b5dfcec4d1b6166067226fae102f7fbcf6bd1bd4',
            'md5': 'd700214df5487801e8ee23d31e60382a',
        }
    )


@skipif_no_network
def test_tararchive_basics(structured_sample_tar_xz: _TestArchive):
    spec = structured_sample_tar_xz
    # this is intentionally a hard-coded POSIX relpath
    member_name = 'test-archive/onetwothree.txt'
    with TarArchiveOperations(spec.path) as archive_ops:
        with archive_ops.open(member_name) as member:
            assert member.read() == spec.content
        with archive_ops.open(PurePosixPath(member_name)) as member:
            assert member.read() == spec.content


@skipif_no_network
def test_tararchive_contain(structured_sample_tar_xz: _TestArchive):
    # this is intentionally a hard-coded POSIX relpath
    member_name = 'test-archive/onetwothree.txt'
    archive_ops = TarArchiveOperations(structured_sample_tar_xz.path)
    # POSIX path str
    assert member_name in archive_ops
    # POSIX path as obj
    assert PurePosixPath(member_name) in archive_ops
    assert 'bogus' not in archive_ops


@skipif_no_network
def test_tararchive_iterator(structured_sample_tar_xz: _TestArchive):
    spec = structured_sample_tar_xz
    with TarArchiveOperations(spec.path) as archive_ops:
        items = list(archive_ops)
        assert len(items) == spec.item_count
        for item in items:
            assert item.name in archive_ops


@skipif_no_network
def test_open(structured_sample_tar_xz: _TestArchive):
    spec = structured_sample_tar_xz
    file_pointer = set()
    with TarArchiveOperations(spec.path) as tf:
        for item in tf:
            if item.type == FileSystemItemType.file:
                with tf.open(str(PurePosixPath(item.name))) as fp:
                    file_pointer.add(fp)
                    assert fp.read(len(spec.content)) == spec.content
        # check the fp before we close the archive handler
        for fp in file_pointer:
            assert fp.closed is True



"""Patch ria_utils.py tp work with abstract RIA-paths

The ORARemote and CreateSiblingRia-patches use an abstract representation of
all paths that are related to elements of a RIA-store, e.g. `ria-layout-version`
or `ria-object-dir`. This patch adapts `ria_utils.py` to this modification.
"""
from __future__ import annotations
import logging
from pathlib import PurePosixPath

from datalad.customremotes.ria_utils import (
    UnknownLayoutVersion,
    get_layout_locations,
)

from . import apply_patch


lgr = logging.getLogger('datalad.customremotes.ria_utils')


# The following two blocks of comments and definitions are verbatim copies from
# `datalad.cutomremotes.ria_utils`

# TODO: Make versions a tuple of (label, description)?
# Object tree versions we introduced so far. This is about the layout within a
# dataset in a RIA store
known_versions_objt = ['1', '2']

# Dataset tree versions we introduced so far. This is about the layout of
# datasets in a RIA store
known_versions_dst = ['1']


# taken from `ria_utils._ensure_version` from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def ria_utils__ensure_version(io, base_path, version):
    """Check a store or dataset version and make sure it is declared

    Parameters
    ----------
    io: SSHRemoteIO or LocalIO
    base_path: PurePosixPath
      root path of a store or dataset
    version: str
      target layout version of the store (dataset tree)
    """
    # PATCH: ensure that `base_path` is an instance of `PurePosixPath`.
    assert isinstance(base_path, PurePosixPath)

    # PATCH: convert abstract `ria-layout-version`-path to concrete IO-specific
    # path
    version_file = io.url2transport_path(base_path / 'ria-layout-version')
    if io.exists(version_file):
        existing_version = io.read_file(version_file).split('|')[0].strip()
        if existing_version != version.split('|')[0]:
            # We have an already existing location with a conflicting version on
            # record.
            # Note, that a config flag after pipe symbol is fine.
            raise ValueError("Conflicting version found at target: {}"
                             .format(existing_version))
        else:
            # already exists, recorded version fits - nothing to do
            return
    # Note, that the following does create the base-path dir as well, since
    # mkdir has parents=True:
    # PATCH: convert abstract path `base_path` to concrete IO-specific path
    # before handing it to `mkdir`.
    io.mkdir(io.url2transport_path(base_path))
    io.write_file(version_file, version)


# taken from `ria_utils.create_store` from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def ria_utils_create_store(io, base_path, version):
    """Helper to create a RIA store

    Note, that this is meant as an internal helper and part of intermediate
    RF'ing. Ultimately should lead to dedicated command or option for
    create-sibling-ria.

    Parameters
    ----------
    io: SSHRemoteIO or LocalIO
      Respective execution instance.
      Note: To be replaced by proper command abstraction
    base_path: PurePosixPath
      root url path of the store
    version: str
      layout version of the store (dataset tree)
    """
    # PATCH: ensure that `base_path` is an instance of `PurePosixPath`.
    assert isinstance(base_path, PurePosixPath)

    # At store level the only version we know as of now is 1.
    if version not in known_versions_dst:
        raise UnknownLayoutVersion("RIA store layout version unknown: {}."
                                   "Supported versions: {}"
                                   .format(version, known_versions_dst))
    _ensure_version(io, base_path, version)
    error_logs = base_path / 'error_logs'
    # PATCH: convert abstract path `error_logs` to concrete IO-specific path
    # before handing it to `mkdir`.
    io.mkdir(io.url2transport_path(error_logs))


# taken from `ria_utils.create_ds_in_store` from datalad-core@864dc4ae24c8aac0ec4003604543b86de4735732
def ria_utils_create_ds_in_store(io,
                                 base_path,
                                 dsid,
                                 obj_version,
                                 store_version,
                                 alias=None,
                                 init_obj_tree=True
                                 ):
    """Helper to create a dataset in a RIA store

    Note, that this is meant as an internal helper and part of intermediate
    RF'ing. Ultimately should lead to a version option for create-sibling-ria
    in conjunction with a store creation command/option.

    Parameters
    ----------
    io: SSHRemoteIO or LocalIO
      Respective execution instance.
      Note: To be replaced by proper command abstraction
    base_path: PurePosixPath
      root path of the store
    dsid: str
      dataset id
    store_version: str
      layout version of the store (dataset tree)
    obj_version: str
      layout version of the dataset itself (object tree)
    alias: str, optional
      alias for the dataset in the store
    init_obj_tree: bool
      whether or not to create the base directory for an annex objects tree (
      'annex/objects')
    """
    # PATCH: ensure that `base_path` is an instance of `PurePosixPath`.
    assert isinstance(base_path, PurePosixPath)

    # TODO: Note for RF'ing, that this is about setting up a valid target
    #       for the special remote not a replacement for create-sibling-ria.
    #       There's currently no git (bare) repo created.

    try:
        # TODO: This is currently store layout version!
        #       Too entangled by current get_layout_locations.
        dsgit_dir, archive_dir, dsobj_dir = \
            get_layout_locations(int(store_version), base_path, dsid)
    except ValueError as e:
        raise UnknownLayoutVersion(str(e))

    if obj_version not in known_versions_objt:
        raise UnknownLayoutVersion("Dataset layout version unknown: {}. "
                                   "Supported: {}"
                                   .format(obj_version, known_versions_objt))

    _ensure_version(io, dsgit_dir, obj_version)

    # PATCH: convert abstract path `archive_dir` to concrete IO-specific path
    # before handing it to `mkdir`.
    io.mkdir(io.url2transport_path(archive_dir))
    if init_obj_tree:
        # PATCH: convert abstract path `dsobj_dir` to concrete IO-specific path
        # before handing it to `mkdir`.
        io.mkdir(io.url2transport_path(dsobj_dir))
    if alias:
        alias_dir = base_path / "alias"
        # PATCH: convert abstract path `alias_dir` to concrete IO-specific path
        # before handing it to `mkdir`.
        io.mkdir(io.url2transport_path(alias_dir))
        try:
            # go for a relative path to keep the alias links valid
            # when moving a store
            io.symlink(
                # PATCH: convert abstract relative path to concrete IO-specific
                # path before handing it to `symlink`.
                io.url2transport_path(
                    PurePosixPath('..') / dsgit_dir.relative_to(base_path)
                ),
                # PATCH: convert abstract alias-path to concrete IO-specific path
                # before handing it to `symlink`.
                io.url2transport_path(alias_dir / alias)
            )
        except FileExistsError:
            lgr.warning("Alias %r already exists in the RIA store, not adding an "
                        "alias.", alias)


_ensure_version = ria_utils__ensure_version


# Overwrite `create_store` to handle paths properly
apply_patch(
    'datalad.customremotes.ria_utils',
    None,
    'create_store',
    ria_utils_create_store,
)


# Overwrite `create_ds_in_store` to handle paths properly
apply_patch(
    'datalad.customremotes.ria_utils',
    None,
    'create_ds_in_store',
    ria_utils_create_ds_in_store,
)


# Overwrite `_ensure_version` to handle paths properly
apply_patch(
    'datalad.customremotes.ria_utils',
    None,
    '_ensure_version',
    ria_utils__ensure_version,
)



from __future__ import annotations


from collections import deque
import logging
from functools import wraps
import os
from os import environ
from pathlib import Path
import pytest
import subprocess
from typing import Any

from datalad.support.external_versions import external_versions
# all datalad-core test utils needed for datalad-next
from datalad.tests.utils_pytest import (
    DEFAULT_BRANCH,
    DEFAULT_REMOTE,
    HTTPPath,
    # TODO REMOVE FOR V2.0
    SkipTest,
    assert_in,
    assert_in_results,
    assert_raises,
    assert_result_count,
    assert_status,
    attr,
    chpwd,
    eq_,
    get_deeply_nested_structure,
    ok_,
    ok_broken_symlink,
    ok_exists,
    ok_good_symlink,
    # TODO REMOVE FOR V2.0
    rmtree,
    skip_if_on_windows,
    skip_if_root,
    skip_wo_symlink_capability,
    swallow_logs,
)
from datalad.tests.test_utils_testrepos import BasicGitTestRepo
from datalad.cli.tests.test_main import run_main
from datalad.ui.progressbars import SilentProgressBar
from datalad.utils import (
    create_tree,
    md5sum,
)
from datalad_next.shell import shell
from datalad_next.utils import (
    CredentialManager,
)

lgr = logging.getLogger("datalad.tests.utils")


class WebDAVPath(object):
    """Serve the content of a path via an HTTP WebDAV URL.

    This class is a context manager.

    Parameters
    ----------
    path : str
        Directory with content to serve.
    auth : tuple
        Username, password

    Returns
    -------
    str
      WebDAV server URL
    """
    def __init__(self, path, auth=None):
        self.path = Path(path)
        self.auth = auth
        self.server = None
        self.server_thread = None

    def __enter__(self):
        try:
            from cheroot import wsgi
            from wsgidav.wsgidav_app import WsgiDAVApp
        except ImportError as e:
            pytest.skip(f'No WSGI capabilities. Install cheroot and/or wsgidav ({e!r})')

        if self.auth:
            auth = {self.auth[0]: {'password': self.auth[1]}}
        else:
            auth = True

        self.path.mkdir(exist_ok=True, parents=True)

        config = {
            "host": "127.0.0.1",
            # random fixed number, maybe make truly random and deal with taken ports
            "port": 43612,
            "provider_mapping": {"/": str(self.path)},
            "simple_dc": {"user_mapping": {'*': auth}},
            # disable DAV server logging to avoid clustering the test output
            # unless logger runs at least on debug log level
            "logging": {"enable": lgr.isEnabledFor(10)},
        }
        app = WsgiDAVApp(config)
        self.server = wsgi.Server(
            bind_addr=(config["host"], config["port"]),
            wsgi_app=app,
        )
        lgr.debug('Starting WebDAV server')
        from threading import Thread
        self.server.prepare()
        self.server_thread = Thread(target=self.server.serve)
        self.server_thread.start()
        lgr.debug('WebDAV started')
        return f'http://{config["host"]}:{config["port"]}'

    def __exit__(self, *args):
        lgr.debug('Stopping WebDAV server')
        # graceful exit
        self.server.stop()
        lgr.debug('WebDAV server stopped, waiting for server thread to exit')
        # wait for shutdown
        self.server_thread.join()
        lgr.debug('WebDAV server thread exited')


def with_credential(name, **kwargs):
    """A decorator to temporarily deploy a credential.

    If a credential of the given name already exists, it will
    be temporarily replaced by the given one.

    In pretty much all cases, the keyword arguments need to include
    `secret`. Otherwise any properties are supported.
    """
    import warnings
    warnings.warn(
        "datalad_next.tests.utils.with_credential was replaced by a `credman` "
        "fixture in datalad_next 1.0, and will be removed in "
        "datalad_next 2.0.",
        DeprecationWarning,
    )

    def with_credential_decorator(fx):
        @wraps(fx)
        def _with_credential(*dargs, **dkw):
            credman = CredentialManager()
            # retrieve anything that might be conflicting with the
            # to-be-deployed credential
            prev_cred = credman.get(name)
            try:
                credman.set(name, **kwargs)
                fx(*dargs, **dkw)
            finally:
                if prev_cred:
                    credman.set(name, **prev_cred)
                else:
                    credman.remove(name)

        return _with_credential
    return with_credential_decorator


def get_git_config_global_fpath() -> Path:
    """Returns the file path for the "global" (aka user) Git config scope"""
    fpath_str = environ.get('GIT_CONFIG_GLOBAL')
    if fpath_str is None:
        # this can happen with the datalad-core setup for Git < 2.32.
        # we provide a fallback, but we do not aim to support all
        # possible variants
        fpath = Path(environ['HOME']) / '.gitconfig'
    else:
        fpath = Path(fpath_str)
    return fpath


class TestUI:
    """Drop-in replacement for the DataLad UI to protocol any calls"""

    is_interactive = False
    """Flag is inspected in generic UI code to check for the possibility
    of interactivity"""

    def __init__(self):
        # this member will hold a log of all calls made to the UI wrapper
        self._log = []

    def __str__(self) -> str:
        return "{cls}(\n{log}\n)".format(
            cls=self.__class__.__name__,
            log='\n'.join(f'  {i[0]}: {i[1]}' for i in self.log),
        )

    @property
    def log(self) -> list:
        """Call log

        Returns
        -------
        list
          Each item is a two-tuple with the label of the UI operation
          as first element, and the verbatim parameters/values of the
          respective operation.
        """
        return self._log

    @property
    def operation_sequence(self) -> list:
        """Same as ``.log()``, but limited to just the operation labels"""
        return [i[0] for i in self.log]

    def question(self, *args, **kwargs) -> Any:
        """Raise ``RuntimeError`` when a question needs to be asked"""
        self._log.append(('question', (args, kwargs)))
        raise RuntimeError(
            'non-interactive test UI was asked for a response to a question')

    def message(self, msg, cr='\n'):
        """Post a message"""
        self._log.append(('message', (msg, cr)))

    def get_progressbar(self, *args, **kwargs):
        """Return a progress handler"""
        self._log.append(('get_progressbar', (args, kwargs)))
        return SilentProgressBar(*args, **kwargs)


class InteractiveTestUI(TestUI):
    """DataLad UI that can also provide staged user responses"""

    is_interactive = True

    def __init__(self):
        super().__init__()
        # queue to provision responses
        self._responses = deque()

    def __str__(self) -> str:
        return "{cls}(\n{log}\n  (unused responses: {res})\n)".format(
            cls=self.__class__.__name__,
            log='\n'.join(f'  {i[0]}: {i[1]}' for i in self.log),
            res=list(self.staged_responses),
        )

    @property
    def staged_responses(self) -> deque:
        """``deque`` for staging user responses and retrieving them"""
        return self._responses

    def question(self, *args, **kwargs) -> Any:
        """Report a provisioned response when a question is asked"""
        self._log.append(('question', (args, kwargs)))
        if not self.staged_responses:
            raise AssertionError(
                "UI response requested, but no further are provisioned")
        response = self.staged_responses.popleft()
        self._log.append(('response', response))
        return response


def assert_ssh_access(
    host: str,
    port: str,
    login: str,
    seckey: str,
    path: str,
    localpath: str | None = None,
):
    """Test for a working SSH connection and sufficient permissions to write

    This helper establishes a connection to an SSH server identified by
    ``host`` and ``port``, using a given SSH private key file (``seckey``) for
    authentication.  Once logged in successfully, it tries to create a
    directory and a file at POSIX ``path`` on the server. If ``localpath`` is
    given, it must be a representation of that server-side path on the local
    file system (e.g., a bindmount), and the helper tests whether the created
    content is also reflected in this directory.
    """
    # we can only handle openssh
    ssh_bin = os.environ.get('DATALAD_SSH_EXECUTABLE', 'ssh')

    ssh_bash_call = [
        ssh_bin,
        '-i', seckey,
        '-p', port,
        f'{login}@{host}',
        'bash',
    ]
    # now try if this is a viable configuration
    # verify execute and write permissions (implicitly also POSIX path handling
    from more_itertools import consume
    with shell(ssh_bash_call) as ssh:
        # each call here will crash with CommandError, if it does not work
        ssh(f'mkdir -p {path}', check=True)
        ssh(f'touch {path}/datalad-tests-probe', check=True)

        if localpath:
            # we should see the probe file locally
            assert (Path(localpath) / 'datalad-tests-probe').exists()

        # cleanup
        ssh(f'rm {path}/datalad-tests-probe', check=True)

    if localpath:
        assert not (Path(localpath) / 'datalad-tests-probe').exists()
